# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import contextlib
import json
import numbers
import time as _time
import typing as t
import xml.etree.ElementTree as etree
from datetime import date, datetime, time
from enum import Enum
from inspect import isclass, isroutine
from types import LambdaType
from urllib.parse import quote

from .._warnings import _warn
from ..partial import Partial
from ..types import PropertyType, _get_taipy_type
from ..utils import (
    _date_to_string,
    _get_broadcast_var_name,
    _get_client_var_name,
    _get_data_type,
    _get_expr_var_name,
    _get_lambda_id,
    _getscopeattr,
    _getscopeattr_drill,
    _is_boolean,
    _is_true,
    _MapDict,
    _to_camel_case,
)
from ..utils.chart_config_builder import _CHART_NAMES, _build_chart_config
from ..utils.table_col_builder import _enhance_columns, _get_name_indexed_property
from ..utils.types import _TaipyBase, _TaipyData, _TaipyToJson
from .json import _TaipyJsonEncoder
from .utils import _add_to_dict_and_get, _get_columns_dict, _get_tuple_val

if t.TYPE_CHECKING:
    from ..gui import Gui


class _Builder:
    """
    Constructs an XML node that can be rendered as a React node.

    This class can only be instantiated internally by Taipy.
    """

    __keys: t.Dict[str, int] = {}

    __BLOCK_CONTROLS = ["dialog", "expandable", "pane", "part"]

    __TABLE_COLUMNS_DEPS = {
        "data",
        "columns",
        "date_format",
        "number_format",
        "nan_value",
        "width",
        "filter",
        "editable",
        "group_by",
        "apply",
        "style",
        "tooltip",
        "lov",
        "row_class_name",
        "cell_class_name",
        "format_fn"
    }

    def __init__(
        self,
        gui: "Gui",
        control_type: str,
        element_name: str,
        attributes: t.Optional[t.Dict[str, t.Any]],
        hash_names: t.Optional[t.Dict[str, str]] = None,
        default_value: t.Optional[t.Any] = "<Empty>",
        lib_name: str = "taipy",
    ):
        from ..gui import Gui
        from .factory import _Factory

        if hash_names is None:
            hash_names = {}

        self.el = etree.Element(element_name)

        self.__control_type = control_type
        self.__element_name = element_name
        self.__lib_name = lib_name
        self.__attributes = attributes or {}
        self.__hashes = hash_names.copy()
        self.__update_vars: t.List[str] = []
        self.__gui: Gui = gui

        self.__default_property_name = _Factory.get_default_property_name(control_type) or ""
        default_property_value = self.__attributes.get(self.__default_property_name, None)
        if default_property_value is None and default_value is not None:
            self.__attributes[self.__default_property_name] = default_value

        # Bind properties dictionary to attributes if condition is matched (will
        # leave the binding for function at the builder )
        if "properties" in self.__attributes:
            (prop_dict, prop_hash) = _Builder.__parse_attribute_value(gui, self.__attributes["properties"])
            if prop_hash is None:
                prop_hash = prop_dict
                prop_hash = self.__gui._bind_var(prop_hash)
                if hasattr(self.__gui._bindings(), prop_hash):
                    prop_dict = _getscopeattr(self.__gui, prop_hash)
            if isinstance(prop_dict, (dict, _MapDict)):
                # Iterate through prop_dict and append to self.attributes
                var_name, _ = gui._get_real_var_name(prop_hash)
                for k, v in prop_dict.items():
                    (val, key_hash) = _Builder.__parse_attribute_value(gui, v)
                    self.__attributes[k] = (
                        f"{{None if ({var_name}) is None else ({var_name}).get('{k}')}}" if key_hash is None else v
                    )
            else:
                _warn(f"{self.__control_type}.properties ({prop_hash}) must be a dict.")

        # Bind potential function and expressions in self.attributes
        self.__hashes.update(_Builder._get_variable_hash_names(gui, self.__attributes, hash_names))

        # set classname
        self.__set_class_names()
        # define a unique key
        self.set_attribute("key", _Builder._get_key(self.__element_name))

    @staticmethod
    def __parse_attribute_value(gui: "Gui", value) -> t.Tuple:
        if isinstance(value, str) and gui._is_expression(value):
            hash_value = gui._evaluate_expr(value)
            try:
                func = gui._get_user_function(hash_value)
                if isroutine(func):
                    return (func, hash_value)
                return (_getscopeattr_drill(gui, hash_value), hash_value)
            except AttributeError:
                _warn(f"Expression '{value}' cannot be evaluated.")
        return (value, None)

    @staticmethod
    def _get_variable_hash_names(
        gui: "Gui", attributes: t.Dict[str, t.Any], hash_names: t.Optional[t.Dict[str, str]] = None
    ) -> t.Dict[str, str]:
        if hash_names is None:
            hash_names = {}
        hashes = {}
        # Bind potential function and expressions in self.attributes
        for k, v in attributes.items():
            hash_name = hash_names.get(k)
            if hash_name is None:
                if isinstance(v, str):
                    # need to unescape the double quotes that were escaped during preprocessing
                    (val, hash_name) = _Builder.__parse_attribute_value(gui, v.replace('\\"', '"'))
                else:
                    val = v
                if isroutine(val) and not hash_name:
                    # if it's not a callable (and not a string), forget it
                    if val.__name__ == "<lambda>":
                        # if it is a lambda and it has already a hash_name, we're fine
                        hash_name = _get_lambda_id(t.cast(LambdaType, val))
                        gui._bind_var_val(hash_name, val)  # type: ignore[arg-type]
                    else:
                        hash_name = _get_expr_var_name(val.__name__)

                if val is not None or hash_name:
                    attributes[k] = val
                if hash_name:
                    hashes[k] = hash_name
        return hashes

    @staticmethod
    def __to_string(x: t.Any) -> str:
        return str(x)

    @staticmethod
    def _get_key(name: str) -> str:
        key_index = _Builder.__keys.get(name, 0)
        _Builder.__keys[name] = key_index + 1
        return f"{name}.{key_index}"

    @staticmethod
    def _reset_key() -> None:
        _Builder.__keys = {}

    def get_name_indexed_property(self, name: str) -> t.Dict[str, t.Any]:
        """
        TODO-undocumented
        Returns all properties defined as <property name>[<named index>] as a dict.

        Arguments:

            name (str): The property name.
        """
        return _get_name_indexed_property(self.__attributes, name)

    def __get_boolean_attribute(self, name: str, default_value=False):
        bool_attr = self.__attributes.get(name, default_value)
        return _is_true(bool_attr) if isinstance(bool_attr, str) else bool(bool_attr)

    def set_boolean_attribute(self, name: str, value: bool):
        """
        TODO-undocumented
        Defines a React Boolean attribute (attr={true|false}).

        Arguments:
            name (str): The property name.
            value (bool): the boolean value.
        """
        return self.__set_react_attribute(_to_camel_case(name), value)

    def set_dict_attribute(self, name: str, default_value: t.Optional[t.Dict[str, t.Any]] = None):
        """
        TODO-undocumented
        Defines a React attribute as a stringified json dict.
        The original property can be a dict or a string formed as <key 1>:<value 1>;<key 2>:<value 2>.

        Arguments:
            name (str): The property name.
            default value (dict): used if no value is specified.
        """
        dict_attr = self.__attributes.get(name)
        if dict_attr is None:
            dict_attr = default_value
        if dict_attr is not None:
            if isinstance(dict_attr, str):
                vals = [x.strip().split(":") for x in dict_attr.split(";")]
                dict_attr = {val[0].strip(): val[1].strip() for val in vals if len(val) > 1}
            if isinstance(dict_attr, (dict, _MapDict)):
                self.__set_json_attribute(_to_camel_case(name), dict_attr)
            else:
                _warn(f"{self.__element_name}: {name} should be a dict: '{str(dict_attr)}'.")
        return self

    def set_dynamic_dict_attribute(self, name: str, default_value: t.Optional[t.Dict[str, t.Any]] = None):
        """
        TODO-undocumented
        Defines a React attribute as a stringified json dict.
        The original property can be a dict or a string formed as <key 1>:<value 1>;<key 2>:<value 2>.

        Arguments:
            name (str): The property name.
            default value (dict): used if no value is specified.
        """
        dict_attr = self.__attributes.get(name)
        if dict_attr is None:
            dict_attr = default_value
        if dict_attr is not None:
            if isinstance(dict_attr, str):
                vals = [x.strip().split(":") for x in dict_attr.split(";")]
                dict_attr = {val[0].strip(): val[1].strip() for val in vals if len(val) > 1}
            if isinstance(dict_attr, (dict, _MapDict)):
                self.__set_json_attribute(_to_camel_case("default_" + name), dict_attr)
            else:
                _warn(f"{self.__element_name}: {name} should be a dict: '{str(dict_attr)}'.")
        if dict_hash := self.__hashes.get(name):
            dict_hash = self.__get_typed_hash_name(dict_hash, PropertyType.dynamic_dict)
            prop_name = _to_camel_case(name)
            self.__update_vars.append(f"{prop_name}={dict_hash}")
            self.__set_react_attribute(prop_name, dict_hash)
        return self

    def __set_json_attribute(self, name, value):
        return self.set_attribute(name, json.dumps(value, cls=_TaipyJsonEncoder))

    def set_number_attribute(self, name: str, default_value: t.Optional[str] = None, optional: t.Optional[bool] = True):
        """
        TODO-undocumented
        Defines a React number attribute (attr={<number>}).

        Arguments:

            name (str): The property name.
            default_value (optional(str)): the default value as a string.
            optional (bool): Default to True, the property is required if False.
        """
        value = self.__attributes.get(name, default_value)
        if value is None:
            if not optional:
                _warn(f"Property {name} is required for control {self.__control_type}.")
            return self
        if isinstance(value, str):
            try:
                val = float(value)
            except ValueError:
                raise ValueError(f"Property {name} expects a number for control {self.__control_type}") from None
        elif isinstance(value, numbers.Number):
            val = value  # type: ignore
        else:
            raise ValueError(
                f"Property {name} expects a number for control {self.__control_type}, received {type(value)}"
            )
        return self.__set_react_attribute(_to_camel_case(name), val)

    def __set_string_attribute(
        self, name: str, default_value: t.Optional[str] = None, optional: t.Optional[bool] = True
    ):
        str_attr = self.__attributes.get(name, default_value)
        if str_attr is None:
            if not optional:
                _warn(f"Property {name} is required for control {self.__control_type}.")
            return self
        return self.set_attribute(_to_camel_case(name), str(str_attr))

    def __set_dynamic_date_attribute(self, var_name: str, default_value: t.Optional[str] = None):
        date_attr = self.__attributes.get(var_name, default_value)
        if date_attr is None:
            date_attr = default_value
        if isinstance(date_attr, (datetime, date, time)):
            value = _date_to_string(date_attr)
            self.set_attribute(_to_camel_case(var_name), value)
        return self

    def __set_dynamic_string_attribute(
        self,
        name: str,
        default_value: t.Optional[str] = None,
        with_update: t.Optional[bool] = False,
        dynamic_property_name: t.Optional[str] = None,
    ):
        str_val = self.__attributes.get(name, default_value)
        if str_val is not None:
            self.set_attribute(
                _to_camel_case(f"default_{name}" if dynamic_property_name is None else name), str(str_val)
            )
        if hash_name := self.__hashes.get(name):
            prop_name = _to_camel_case(name if dynamic_property_name is None else dynamic_property_name)
            if with_update:
                self.__update_vars.append(f"{prop_name}={hash_name}")
            self.__set_react_attribute(prop_name, hash_name)
        return self

    def __set_function_attribute(
        self, name: str, default_value: t.Optional[str] = None, optional: t.Optional[bool] = True
    ):
        str_attr = self.__attributes.get(name, default_value)
        if str_attr is None:
            if not optional:
                _warn(f"Property {name} is required for control {self.__control_type}.")
            return self
        elif isroutine(str_attr):
            str_attr = self.__hashes.get(name)
            if str_attr is None:
                return self
        elif _is_boolean(str_attr) and not _is_true(t.cast(str, str_attr)):
            return self.__set_react_attribute(_to_camel_case(name), False)
        elif str_attr:
            str_attr = str(str_attr)
            func = self.__gui._get_user_function(str_attr)
            if func == str_attr:
                _warn(f"{self.__control_type}.{name}: {str_attr} is not a function.")
        return self.set_attribute(_to_camel_case(name), str_attr) if str_attr else self

    def __set_string_or_number_attribute(self, name: str, default_value: t.Optional[t.Any] = None):
        attr = self.__attributes.get(name, default_value)
        if attr is None:
            return self
        if isinstance(attr, numbers.Number):
            return self.__set_react_attribute(_to_camel_case(name), attr)
        else:
            return self.set_attribute(_to_camel_case(name), attr)

    def __set_react_attribute(self, name: str, value: t.Any):
        return self.set_attribute(name, "{!" + (str(value).lower() if isinstance(value, bool) else str(value)) + "!}")

    @staticmethod
    def enum_adapter(e: t.Union[Enum, str]):
        return (e.value, e.name) if isinstance(e, Enum) else e

    def _get_lov_adapter(  # noqa: C901
        self, var_name: str, property_name: t.Optional[str] = None, multi_selection=True, with_default=True
    ):
        property_name = var_name if property_name is None else property_name
        lov_name = self.__hashes.get(var_name)
        real_var_name = self.__gui._get_real_var_name(lov_name)[0] if lov_name else None
        lov = self.__attributes.get(var_name)
        adapter: t.Any = None
        var_type: t.Optional[str] = None
        if isinstance(lov, str):
            lov = list(lov.split(";"))
        if isclass(lov) and issubclass(lov, Enum):
            adapter = _Builder.enum_adapter
            var_type = "Enum"
            lov = list(lov)
        if not isinstance(lov, list) and hasattr(lov, "tolist"):
            try:
                lov = lov.tolist()  # type: ignore[union-attr]
            except Exception as e:
                _warn(f"Error accessing List of values for '{real_var_name or property_name}'", e)
                lov = None

        default_lov: t.Optional[t.List[t.Any]] = [] if with_default or not lov_name else None

        adapter = self.__attributes.get("adapter", adapter)
        if adapter and isinstance(adapter, str):
            adapter = self.__gui._get_user_function(adapter)
        if adapter and not callable(adapter):
            _warn(f"{self.__element_name}: adapter property value is invalid.")
            adapter = None
        var_type = self.__attributes.get("type", var_type)
        if isclass(var_type):
            var_type = var_type.__name__  # type: ignore

        if isinstance(lov, list):
            if not isinstance(var_type, str):
                elt = None
                if lov:
                    elt = lov[0]
                else:
                    value = self.__attributes.get("value")
                    if isinstance(value, list):
                        if len(value) > 0:
                            elt = value[0]
                    else:
                        elt = value
                var_type = self.__gui._get_unique_type_adapter(type(elt).__name__)
            if adapter is None:
                adapter = self.__gui._get_adapter_for_type(var_type)
            elif var_type == str.__name__ and isroutine(adapter):
                var_type += (
                    _get_lambda_id(t.cast(LambdaType, adapter))
                    if adapter.__name__ == "<lambda>"
                    else _get_expr_var_name(adapter.__name__)
                )
            if lov_name:
                if adapter is None:
                    adapter = self.__gui._get_adapter_for_type(lov_name)
                else:
                    self.__gui._add_type_for_var(lov_name, t.cast(str, var_type))
            if value_name := self.__hashes.get("value"):
                if adapter is None:
                    adapter = self.__gui._get_adapter_for_type(value_name)
                else:
                    self.__gui._add_type_for_var(value_name, t.cast(str, var_type))
            if adapter is not None:
                self.__gui._add_adapter_for_type(var_type, adapter)  # type: ignore

            if default_lov is not None and lov:
                for elt in lov:
                    ret = self.__gui._run_adapter(
                        t.cast(t.Callable, adapter), elt, adapter.__name__ if isroutine(adapter) else "adapter"
                    )  # type: ignore
                    if ret is not None:
                        default_lov.append(ret)

            ret_list = []
            value = self.__attributes.get("value")
            val_list = value if isinstance(value, list) else [value]
            for val in val_list:
                ret = self.__gui._run_adapter(
                    t.cast(t.Callable, adapter),
                    val,
                    adapter.__name__ if isroutine(adapter) else "adapter",
                    id_only=True,
                )  # type: ignore
                if ret is not None:
                    ret_list.append(ret)
            if multi_selection:
                self.__set_default_value("value", ret_list)
            else:
                ret_val = ret_list[0] if len(ret_list) else ""
                if ret_val == "-1" and self.__attributes.get("unselected_value") is not None:
                    ret_val = str(self.__attributes.get("unselected_value", ""))
                self.__set_default_value("value", ret_val)

        # LoV default value
        if default_lov is not None:
            self.__set_json_attribute(_to_camel_case(f"default_{property_name}"), default_lov)

        # LoV expression binding
        if lov_name and real_var_name:
            typed_lov_hash = (
                self.__gui._evaluate_expr(
                    f"{{{self.__gui._get_call_method_name('_get_adapted_lov')}({real_var_name},'{var_type}')}}"
                )
                if var_type
                else lov_name
            )
            hash_name = self.__get_typed_hash_name(typed_lov_hash, PropertyType.lov)
            camel_prop = _to_camel_case(property_name)
            self.__update_vars.append(f"{camel_prop}={hash_name}")
            self.__set_react_attribute(camel_prop, hash_name)

        return self

    def __filter_attribute_names(self, names: t.Iterable[str]):
        return [k for k in self.__attributes if k in names or any(k.startswith(n + "[") for n in names)]

    def __get_held_name(self, key: str):
        name = self.__hashes.get(key)
        if name:
            v = self.__attributes.get(key)
            if isinstance(v, _TaipyBase):
                return name[: len(v.get_hash()) + 1]
        return name

    def __filter_attributes_hashes(self, keys: t.List[str]):
        hash_names = [k for k in self.__hashes if k in keys]
        attr_names = [k for k in keys if k not in hash_names]
        return (
            {k: v for k, v in self.__attributes.items() if k in attr_names},
            {k: self.__get_held_name(k) for k in self.__hashes if k in hash_names},
        )

    def __build_rebuild_fn(self, fn_name: str, attribute_names: t.Iterable[str]):
        rebuild = self.__attributes.get("rebuild", False)
        rebuild_hash = self.__hashes.get("rebuild")
        if rebuild_hash or _is_true(rebuild):
            attributes, hashes = self.__filter_attributes_hashes(self.__filter_attribute_names(attribute_names))
            rebuild_name = f"bool({self.__gui._get_real_var_name(rebuild_hash)[0]})" if rebuild_hash else "None"
            try:
                self.__gui._set_building(True)
                return self.__gui._evaluate_expr(
                    "{"
                    + f'{fn_name}({rebuild}, {rebuild_name}, "{quote(json.dumps(attributes))}", "{quote(json.dumps(hashes))}", {", ".join([f"{k}={v2}" for k, v2 in {v: self.__gui._get_real_var_name(t.cast(str, v))[0] for v in hashes.values()}.items()])})'  # noqa: E501
                    + "}"
                )
            finally:
                self.__gui._set_building(False)
        return None

    def _get_dataframe_attributes(self) -> "_Builder":
        date_format = _add_to_dict_and_get(self.__attributes, "date_format", "MM/dd/yyyy")
        data = self.__attributes.get("data")
        data_hash = self.__hashes.get("data", "")
        cmp_hash = ""
        if data_hash:
            cmp_idx = 1
            cmp_datas = []
            cmp_datas_hash = []
            while cmp_data := self.__hashes.get(f"data[{cmp_idx}]"):
                cmp_idx += 1
                cmp_datas.append(self.__gui._get_real_var_name(cmp_data)[0])
                cmp_datas_hash.append(cmp_data)
            if cmp_datas:
                cmp_hash = self.__gui._evaluate_expr(
                    "{"
                    + f"{self.__gui._get_call_method_name('_compare_data')}"
                    + f'({self.__gui._get_real_var_name(data_hash)[0]},{",".join(cmp_datas)})'
                    + "}"
                )
                self.__update_vars.append(f"comparedatas={','.join(cmp_datas_hash)}")
        col_types = self.__gui._get_accessor().get_col_types(data_hash, _TaipyData(data, data_hash))
        col_dict = _get_columns_dict(
            data, self.__attributes.get("columns", {}), col_types, date_format, self.__attributes.get("number_format")
        )

        rebuild_fn_hash = self.__build_rebuild_fn(
            self.__gui._get_call_method_name("_tbl_cols"), _Builder.__TABLE_COLUMNS_DEPS
        )
        if rebuild_fn_hash:
            self.__set_react_attribute("columns", rebuild_fn_hash)
        if col_dict is not None:
            _enhance_columns(self.__attributes, self.__hashes, col_dict, self.__element_name)
            self.__set_json_attribute("defaultColumns", col_dict)
        if cmp_hash:
            hash_name = self.__get_typed_hash_name(cmp_hash, PropertyType.data)
            self.__set_react_attribute(
                _to_camel_case("data"),
                _get_client_var_name(hash_name),
            )
            self.__set_update_var_name(hash_name)
            self.set_boolean_attribute("compare", True)
            self.__set_string_attribute("on_compare")

        if not isinstance(self.__attributes.get("style"), (type(None), dict, _MapDict)):
            _warn("Table: property 'style' has been renamed to 'row_class_name'.")
        if row_class_name := self.__attributes.get("row_class_name"):
            if isroutine(row_class_name):
                value = self.__hashes.get("row_class_name")
            elif isinstance(row_class_name, str):
                value = row_class_name.strip()
            else:
                value = None
            if value in col_types.keys():
                _warn(f"{self.__element_name}: row_class_name={value} must not be a column name.")
            elif value:
                self.set_attribute("rowClassName", value)
        if tooltip := self.__attributes.get("tooltip"):
            if isroutine(tooltip):
                value = self.__hashes.get("tooltip")
            elif isinstance(tooltip, str):
                value = tooltip.strip()
            else:
                value = None
            if value in col_types.keys():
                _warn(f"{self.__element_name}: tooltip={value} must not be a column name.")
            elif value:
                self.set_attribute("tooltip", value)
        return self

    def _get_chart_config(self, default_type: str, default_mode: str):
        self.__attributes["_default_type"] = default_type
        self.__attributes["_default_mode"] = default_mode
        rebuild_fn_hash = self.__build_rebuild_fn(
            self.__gui._get_call_method_name("_chart_conf"),
            _CHART_NAMES + ("_default_type", "_default_mode", "data"),
        )
        if rebuild_fn_hash:
            self.__set_react_attribute("config", rebuild_fn_hash)

        # read column definitions
        data = self.__attributes.get("data")
        data_hash = self.__hashes.get("data", "")
        col_types = self.__gui._get_accessor().get_col_types(data_hash, _TaipyData(data, data_hash))

        config = _build_chart_config(self.__gui, self.__attributes, col_types)

        self.__set_json_attribute("defaultConfig", config)
        self._set_chart_selected(max=len(config.get("traces", [])))
        self.__set_refresh_on_update()
        return self

    def _set_string_with_check(self, var_name: str, values: t.List[str], default_value: t.Optional[str] = None):
        value = self.__attributes.get(var_name, default_value)
        if value is not None:
            value = str(value).lower()
            self.__attributes[var_name] = value
            if value not in values:
                _warn(f"{self.__element_name}: {var_name}={value} should be in {values}.")
            else:
                self.__set_string_attribute(var_name, default_value)
        return self

    def __set_list_attribute(
        self,
        name: str,
        hash_name: t.Optional[str],
        val: t.Any,
        elt_type: t.Type,
        dynamic=True,
        default_val: t.Optional[t.Any] = None,
    ) -> t.List[str]:
        val = default_val if val is None else val
        if not hash_name and isinstance(val, str):
            val = [elt_type(t.strip()) for t in val.split(";")]
        if isinstance(val, list):
            if hash_name and dynamic:
                self.__set_react_attribute(name, hash_name)
                return [f"{name}={hash_name}"]
            else:
                self.__set_json_attribute(name, val)
        elif val is not None:
            _warn(f"{self.__element_name}: {name} should be a list of {elt_type}.")
        return []

    def _set_chart_selected(self, max=0):
        name = "selected"
        default_sel = self.__attributes.get(name)
        if not isinstance(default_sel, list) and name in self.__attributes:
            default_sel = []
        if max == 0:
            self.__update_vars.extend(
                self.__set_list_attribute(
                    name,
                    self.__hashes.get(name),
                    default_sel,
                    int,
                )
            )
            return
        idx = 1
        name_idx = f"{name}[{idx}]"
        sel = self.__attributes.get(name_idx)
        if not isinstance(sel, list) and name_idx in self.__attributes:
            sel = []
        while idx <= max or name_idx in self.__attributes:
            if sel is not None or default_sel is not None:
                self.__update_vars.extend(
                    self.__set_list_attribute(
                        f"{name}{idx - 1}",
                        self.__hashes.get(name_idx if sel is not None else name),
                        sel if sel is not None else default_sel,
                        int,
                    )
                )
            idx += 1
            name_idx = f"{name}[{idx}]"
            sel = self.__attributes.get(name_idx)
            if not isinstance(sel, list) and name_idx in self.__attributes:
                sel = []

    def _get_list_attribute(self, name: str, list_type: PropertyType):
        hash_name = self.__hashes.get(name)
        if hash_name is None:
            list_val = self.__attributes.get(name)
            if isinstance(list_val, str):
                list_val = list(list_val.split(";"))
            if isinstance(list_val, list):
                # TODO catch the cast exception
                if list_type is PropertyType.number:
                    list_val = [int(v) for v in list_val]
                else:
                    list_val = [str(v) for v in list_val]
                if list_val:
                    self.__set_react_attribute(_to_camel_case(name), list_val)
            elif list_val is not None:
                    _warn(f"{self.__element_name}: {name} should be a list.")
        else:
            self.__set_react_attribute(_to_camel_case(name), hash_name)
            self.__update_vars.append(f"{_to_camel_case(name)}={hash_name}")
        return self

    def __set_class_names(self):
        self.set_attribute("libClassName", self.__lib_name + "-" + self.__control_type.replace("_", "-"))
        return self.__set_dynamic_string_attribute("class_name", dynamic_property_name="dynamic_class_name")

    def _set_dataType(self):
        value = self.__attributes.get("value")
        return self.set_attribute("dataType", _get_data_type(value))

    def _set_file_content(self, var_name: str = "content"):
        if hash_name := self.__hashes.get(var_name):
            self.__set_update_var_name(hash_name)
        else:
            _warn(f"file_selector: {var_name} should be bound.")
        return self

    def _set_content(self, var_name: str = "content", image=True):
        content = self.__attributes.get(var_name)
        hash_name = self.__hashes.get(var_name)
        if content is None and hash_name is None:
            return self
        value = self.__gui._get_content(hash_name or var_name, content, image)
        if hash_name:
            hash_name = self.__get_typed_hash_name(hash_name, PropertyType.image if image else PropertyType.content)
        if hash_name:
            self.__set_react_attribute(
                var_name,
                _get_client_var_name(hash_name),
            )
        return self.set_attribute(_to_camel_case(f"default_{var_name}"), value)

    def __set_dynamic_string_list(self, var_name: str, default_value: t.Any):
        hash_name = self.__hashes.get(var_name)
        loi = self.__attributes.get(var_name)
        if loi is None:
            loi = default_value
        if isinstance(loi, str):
            loi = [s.strip() for s in loi.split(";") if s.strip()]
        if isinstance(loi, list):
            self.__set_json_attribute(_to_camel_case(f"default_{var_name}"), loi)
        if hash_name:
            self.__update_vars.append(f"{var_name}={hash_name}")
            self.__set_react_attribute(var_name, hash_name)
        return self

    def __set_dynamic_number_attribute(self, var_name: str, default_value: t.Any):
        hash_name = self.__hashes.get(var_name)
        numVal = self.__attributes.get(var_name)
        if numVal is None:
            numVal = default_value
        if isinstance(numVal, str):
            try:
                numVal = float(numVal)
            except Exception as e:
                _warn(f"{self.__element_name}: {var_name} cannot be transformed into a number", e)
                numVal = 0
        if isinstance(numVal, numbers.Number):
            self.__set_react_attribute(_to_camel_case(f"default_{var_name}"), numVal)
        elif numVal is not None:
            _warn(f"{self.__element_name}: {var_name} value is not valid ({numVal}).")
        if hash_name:
            hash_name = self.__get_typed_hash_name(hash_name, PropertyType.number)
            self.__update_vars.append(f"{var_name}={hash_name}")
            self.__set_react_attribute(var_name, hash_name)
        return self

    def __set_default_value(
        self,
        var_name: str,
        value: t.Optional[t.Any] = None,
        native_type: bool = False,
        var_type: t.Optional[PropertyType] = None,
    ):
        if value is None:
            value = self.__attributes.get(var_name)
        default_var_name = _to_camel_case(f"default_{var_name}")
        if isinstance(value, (datetime, date, time)):
            return self.set_attribute(default_var_name, _date_to_string(value))
        elif isinstance(value, str):
            return self.set_attribute(default_var_name, value)
        elif native_type and isinstance(value, numbers.Number):
            return self.__set_react_attribute(default_var_name, value)
        elif value is None:
            return self.__set_react_attribute(default_var_name, "null")
        elif var_type == PropertyType.lov_value:
            # Done by _get_adapter
            return self
        elif isclass(var_type) and issubclass(var_type, _TaipyBase):  # type: ignore
            return self.__set_default_value(var_name, t.cast(t.Callable, var_type)(value, "").get())
        else:
            return self.__set_json_attribute(default_var_name, value)

    def __set_update_var_name(self, hash_name: str):
        return self.set_attribute("updateVarName", hash_name)

    def set_value_and_default(
        self,
        var_name: t.Optional[str] = None,
        with_update=True,
        with_default=True,
        native_type=False,
        var_type: t.Optional[PropertyType] = None,
        default_val: t.Any = None,
    ):
        """
        TODO-undocumented
        Sets the value associated with the default property.

        Arguments:

            var_name (str): The property name (default to default property name).
            with_update (optional(bool)): Should the attribute be dynamic (default True).
            with_default (optional(bool)): Should a default attribute be set (default True).
            native_type (optional(bool)): If var_type == dynamic_number, parse the value to number.
            var_type (optional(PropertyType)): the property type (default to string).
            default_val (optional(Any)): the default value.
        """
        var_name = self.__default_property_name if var_name is None else var_name
        if var_type == PropertyType.slider_value or var_type == PropertyType.toggle_value:
            if self.__attributes.get("lov"):
                var_type = PropertyType.lov_value
                native_type = False
            elif var_type == PropertyType.toggle_value:
                self.__set_react_attribute(_to_camel_case("is_switch"), True)
                var_type = PropertyType.dynamic_boolean
                native_type = True
            else:
                var_type = (
                    PropertyType.dynamic_lo_numbers
                    if isinstance(self.__attributes.get("value"), list)
                    else PropertyType.dynamic_number
                )
                native_type = True
        if var_type == PropertyType.dynamic_boolean:
            return self.set_attributes([(var_name, var_type, bool(default_val), with_update)])
        if hash_name := self.__hashes.get(var_name):
            hash_name = self.__get_typed_hash_name(hash_name, var_type)
            self.__set_react_attribute(
                _to_camel_case(var_name),
                _get_client_var_name(hash_name),
            )
            if with_update:
                self.__set_update_var_name(hash_name)
            if with_default:
                if native_type:
                    val = self.__attributes.get(var_name)
                    if native_type and isinstance(val, str):
                        with contextlib.suppress(Exception):
                            val = float(val)
                    self.__set_default_value(var_name, val, native_type=native_type)
                else:
                    self.__set_default_value(var_name, var_type=var_type)
        else:
            if var_type == PropertyType.data and (self.__control_type != "chart" or "figure" not in self.__attributes):
                _warn(f"{self.__control_type}.{var_name} property should be bound.")
            value = self.__attributes.get(var_name)
            if value is not None:
                if native_type:
                    if isinstance(value, str):
                        with contextlib.suppress(Exception):
                            value = float(value)
                    if isinstance(value, (int, float)):
                        return self.__set_react_attribute(_to_camel_case(var_name), value)
                if isinstance(value, (datetime, date, time)):
                    value = _date_to_string(value)
                self.set_attribute(_to_camel_case(var_name), value)
        return self

    def _set_labels(self, var_name: str = "labels"):
        if value := self.__attributes.get(var_name):
            if _is_true(value):
                return self.__set_react_attribute(_to_camel_case(var_name), True)
            elif isinstance(value, (dict, _MapDict)):
                return self.set_dict_attribute(var_name)
        return self

    def _set_partial(self):
        if self.__control_type not in _Builder.__BLOCK_CONTROLS:
            return self
        if partial := self.__attributes.get("partial"):
            if self.__attributes.get("page"):
                _warn(f"{self.__element_name} control: page and partial should not be both defined.")
            if isinstance(partial, Partial):
                self.__attributes["page"] = partial._route
                self.__set_react_attribute("partial", partial._route)
                self.__set_react_attribute("defaultPartial", True)
        return self

    def _set_propagate(self):
        val = self.__get_boolean_attribute("propagate", t.cast(bool, self.__gui._config.config.get("propagate")))
        return self if val else self.set_boolean_attribute("propagate", False)

    def __set_refresh_on_update(self):
        if self.__update_vars:
            self.set_attribute("updateVars", ";".join(self.__update_vars))
        return self

    def _set_table_pagesize_options(self, default_size=None):
        if default_size is None:
            default_size = [50, 100, 500]
        page_size_options = self.__attributes.get("page_size_options", default_size)
        if isinstance(page_size_options, str):
            try:
                page_size_options = [int(s.strip()) for s in page_size_options.split(";")]
            except Exception as e:
                _warn(f"{self.__element_name}: page_size_options value is invalid ({page_size_options})", e)
        if isinstance(page_size_options, list):
            self.__set_json_attribute("pageSizeOptions", page_size_options)
        else:
            _warn(f"{self.__element_name}: page_size_options should be a list.")
        return self

    def _set_input_type(self, type_name: str, allow_password=False):
        if allow_password and self.__get_boolean_attribute("password", False):
            return self.set_attribute("type", "password")
        return self.set_attribute("type", self.__attributes.get("type", type_name))

    def _set_kind(self):
        if self.__attributes.get("theme", False):
            self.set_attribute("mode", "theme")
        return self

    def __get_typed_hash_name(self, hash_name: str, var_type: t.Optional[PropertyType]) -> str:
        if taipy_type := _get_taipy_type(var_type):
            expr = self.__gui._get_expr_from_hash(hash_name)
            hash_name = self.__gui._evaluate_bind_holder(t.cast(t.Type[_TaipyBase], taipy_type), expr)
        return hash_name

    def __set_dynamic_bool_attribute(self, name: str, def_val: t.Any, with_update: bool, update_main=True):
        hash_name = self.__hashes.get(name)
        val = self.__get_boolean_attribute(name, def_val)
        default_name = f"default_{name}" if hash_name is not None else name
        if val != def_val:
            self.set_boolean_attribute(default_name, val)
        if hash_name is not None:
            hash_name = self.__get_typed_hash_name(hash_name, PropertyType.dynamic_boolean)
            self.__set_react_attribute(_to_camel_case(name), _get_client_var_name(hash_name))
            if with_update:
                if update_main:
                    self.__set_update_var_name(hash_name)
                else:
                    self.__update_vars.append(f"{_to_camel_case(name)}={hash_name}")

    def __set_dynamic_property_without_default(
        self, name: str, property_type: PropertyType, optional: t.Optional[bool] = False
    ):
        hash_name = self.__hashes.get(name)
        if hash_name is None:
            if not optional:
                _warn(f"{self.__element_name}.{name} should be bound.")
        else:
            hash_name = self.__get_typed_hash_name(hash_name, property_type)
            self.__update_vars.append(f"{_to_camel_case(name)}={hash_name}")
            self.__set_react_attribute(_to_camel_case(name), _get_client_var_name(hash_name))
        return self

    def __set_html_content(self, name: str, property_name: str, property_type: PropertyType):
        hash_name = self.__hashes.get(name)
        if not hash_name:
            return self
        front_var = self.__get_typed_hash_name(hash_name, property_type)
        self.set_attribute(
            _to_camel_case(f"default_{property_name}"),
            self.__gui._get_user_content_url(
                None,
                {
                    "variable_name": front_var,
                    self.__gui._HTML_CONTENT_KEY: str(_time.time()),
                },
            ),
        )
        return self.__set_react_attribute(_to_camel_case(property_name), _get_client_var_name(front_var))

    def set_attributes(self, attributes: t.List[tuple]):  # noqa: C901
        """
        TODO-undocumented
        Sets the attributes from the property with type and default value.

        Arguments:

            attributes (list(tuple)): The list of attributes as (property name, property type, default value).
        """
        attributes.append(("id",))  # Every element should have an id attribute
        for attr in attributes:
            if not isinstance(attr, tuple):
                attr = (attr,)
            var_type = _get_tuple_val(attr, 1, PropertyType.string)
            if var_type == PropertyType.to_json:
                var_type = _TaipyToJson
            if var_type == PropertyType.boolean:
                def_val = _get_tuple_val(attr, 2, False)
                if isinstance(def_val, bool) or self.__attributes.get(attr[0], None) is not None:
                    val = self.__get_boolean_attribute(attr[0], def_val)
                    if val != def_val:
                        self.set_boolean_attribute(attr[0], val)
            elif var_type == PropertyType.dynamic_boolean:
                self.__set_dynamic_bool_attribute(
                    attr[0],
                    _get_tuple_val(attr, 2, False),
                    _get_tuple_val(attr, 3, False),
                    _get_tuple_val(attr, 4, True),
                )
            elif var_type == PropertyType.number:
                self.set_number_attribute(attr[0], _get_tuple_val(attr, 2, None))
            elif var_type == PropertyType.dynamic_number:
                self.__set_dynamic_number_attribute(attr[0], _get_tuple_val(attr, 2, None))
            elif var_type == PropertyType.string:
                self.__set_string_attribute(attr[0], _get_tuple_val(attr, 2, None), _get_tuple_val(attr, 3, True))
            elif var_type == PropertyType.dynamic_string:
                self.__set_dynamic_string_attribute(
                    attr[0], _get_tuple_val(attr, 2, None), _get_tuple_val(attr, 3, False)
                )
            elif var_type == PropertyType.string_list:
                self.__update_vars.extend(
                    self.__set_list_attribute(
                        attr[0],
                        self.__hashes.get(attr[0]),
                        self.__attributes.get(attr[0]),
                        str,
                        False,
                        _get_tuple_val(attr, 2, None),
                    )
                )
            elif var_type == PropertyType.function:
                self.__set_function_attribute(attr[0], _get_tuple_val(attr, 2, None), _get_tuple_val(attr, 3, True))
            elif var_type == PropertyType.react:
                prop_name = _to_camel_case(attr[0])
                if hash_name := self.__hashes.get(attr[0]):
                    self.__update_vars.append(f"{prop_name}={hash_name}")
                    self.__set_react_attribute(prop_name, hash_name)
                else:
                    self.__set_react_attribute(prop_name, self.__attributes.get(attr[0], _get_tuple_val(attr, 2, None)))
            elif var_type == PropertyType.broadcast:
                self.__set_react_attribute(
                    _to_camel_case(attr[0]), _get_broadcast_var_name(_get_tuple_val(attr, 2, None))
                )
            elif var_type == PropertyType.string_or_number:
                self.__set_string_or_number_attribute(attr[0], _get_tuple_val(attr, 2, None))
            elif var_type == PropertyType.dict:
                self.set_dict_attribute(attr[0], _get_tuple_val(attr, 2, None))
            elif var_type == PropertyType.dynamic_dict:
                self.set_dynamic_dict_attribute(attr[0], _get_tuple_val(attr, 2, None))
            elif var_type == PropertyType.dynamic_list:
                self.__set_dynamic_string_list(attr[0], _get_tuple_val(attr, 2, None))
            elif var_type == PropertyType.boolean_or_list:
                if _is_boolean(self.__attributes.get(attr[0])):
                    self.__set_dynamic_bool_attribute(attr[0], _get_tuple_val(attr, 2, False), True, update_main=False)
                else:
                    self.__set_dynamic_string_list(attr[0], _get_tuple_val(attr, 2, None))
            elif var_type == PropertyType.dynamic_date:
                self.__set_dynamic_date_attribute(attr[0], _get_tuple_val(attr, 2, None))
            elif var_type == PropertyType.data:
                self.__set_dynamic_property_without_default(attr[0], t.cast(PropertyType, var_type))
            elif (
                var_type == PropertyType.lov
                or var_type == PropertyType.single_lov
                or var_type == PropertyType.lov_no_default
            ):
                self._get_lov_adapter(
                    attr[0],
                    multi_selection=var_type != PropertyType.single_lov,
                    with_default=var_type != PropertyType.lov_no_default,
                )
            elif var_type == PropertyType.lov_value:
                self.__set_dynamic_property_without_default(
                    attr[0], t.cast(PropertyType, var_type), _get_tuple_val(attr, 2, None) == "optional"
                )
            elif var_type == PropertyType.toHtmlContent:
                self.__set_html_content(attr[0], "page", t.cast(PropertyType, var_type))
            elif isclass(var_type) and issubclass(var_type, _TaipyBase):
                prop_name = _to_camel_case(attr[0])
                if hash_name := self.__hashes.get(attr[0]):
                    expr = self.__gui._get_expr_from_hash(hash_name)
                    hash_name = self.__gui._evaluate_bind_holder(var_type, expr)
                    self.__update_vars.append(f"{prop_name}={hash_name}")
                    self.__set_react_attribute(prop_name, hash_name)
                else:
                    val = self.__attributes.get(attr[0])
                    self.set_attribute(
                        prop_name, var_type(_get_tuple_val(attr, 2, None) if val is None else val, "").get()
                    )

        self.__set_refresh_on_update()
        return self

    def set_attribute(self, name: str, value: t.Any):
        """
        TODO-undocumented
        Sets an attribute.

        Arguments:

            name (str): The name of the attribute.
            value (Any): The value of the attribute (must be json serializable).
        """
        if value is not None:
            self.el.set(name, value)
        return self

    def get_element(self):
        """
        TODO-undocumented
        Returns the xml.etree.ElementTree.Element
        """
        return self.el

    def _build_to_string(self):
        el_str = str(etree.tostring(self.el, encoding="utf8").decode("utf8"))
        el_str = el_str.replace("<?xml version='1.0' encoding='utf8'?>\n", "")
        el_str = el_str.replace("/>", ">")
        return el_str, self.__element_name



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
from __future__ import annotations

import typing as t
from abc import ABC, abstractmethod
from datetime import date, datetime, time, timedelta
from json import JSONEncoder
from pathlib import Path

import numpy
import pandas
from flask.json.provider import DefaultJSONProvider

from .._warnings import _warn
from ..icon import Icon
from ..utils import _date_to_string, _DoNotUpdate, _MapDict, _TaipyBase
from ..utils.singleton import _Singleton


class JsonAdapter(ABC):
    """NOT DOCUMENTED"""
    def register(self):
        _TaipyJsonAdapter().register(self)

    @abstractmethod
    def parse(self, o) -> t.Union[t.Any, None]:
        return None


class _DefaultJsonAdapter(JsonAdapter):
    def parse(self, o):
        if isinstance(o, Icon):
            return o._to_dict()
        if isinstance(o, _MapDict):
            return o._dict
        if isinstance(o, _TaipyBase):
            return o.get()
        if isinstance(o, (datetime, date, time)):
            return _date_to_string(o)
        if isinstance(o, Path):
            return str(o)
        if isinstance(o, (timedelta, pandas.Timedelta)):
            return str(o)
        if isinstance(o, numpy.generic):
            return getattr(o, "tolist", lambda: o)()
        if isinstance(o, _DoNotUpdate):
            return None


class _TaipyJsonAdapter(object, metaclass=_Singleton):
    def __init__(self) -> None:
        self._adapters: t.List[JsonAdapter] = []
        self.register(_DefaultJsonAdapter())

    def register(self, adapter: JsonAdapter):
        self._adapters.append(adapter)

    def parse(self, o):
        try:
            for adapter in reversed(self._adapters):
                if (output := adapter.parse(o)) is not None:
                    return output
            raise TypeError(f"Object of type {type(o).__name__} is not JSON serializable (value: {o}).")
        except Exception as e:
            _warn("Exception while resolving JSON", e)
            return None


class _TaipyJsonEncoder(JSONEncoder):
    def default(self, o):
        return _TaipyJsonAdapter().parse(o)


class _TaipyJsonProvider(DefaultJSONProvider):
    default = staticmethod(_TaipyJsonAdapter().parse)  # type: ignore
    sort_keys = False



import typing as t
from pathlib import Path

from _typeshed import Incomplete
from flask import Flask

from .config import Config
from .extension.library import ElementLibrary
from .page import Page
from .partial import Partial
from .state import State

class Gui:
    on_action: t.Callable | None
    on_change: t.Callable | None
    on_init: t.Callable | None
    on_page_load: t.Callable | None
    on_navigate: t.Callable | None
    on_exception: t.Callable | None
    on_status: t.Callable | None
    on_user_content: t.Callable | None
    def __init__(
        self,
        page: str | Page | None = None,
        pages: dict | None = None,
        css_file: str | None = None,
        path_mapping: dict | None = None,
        env_filename: str | None = None,
        libraries: list[ElementLibrary] | None = None,
        flask: Flask | None = None,
    ) -> None: ...
    @staticmethod
    def add_library(library: ElementLibrary) -> None: ...
    @staticmethod
    def register_content_provider(
        content_type: type, content_provider: t.Callable[..., str]
    ) -> None: ...
    @staticmethod
    def add_shared_variable(*names: str) -> None: ...
    @staticmethod
    def add_shared_variables(*names: str) -> None: ...
    def __enter__(self): ...
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ): ...
    def invoke_callback(
        self,
        state_id: str,
        callback: t.Callable | str,
        args: t.Sequence[t.Any] | None = None,
        module_context: str | None = None,
    ) -> t.Any: ...
    def broadcast_callback(
        self,
        callback: t.Callable,
        args: t.Sequence[t.Any] | None = None,
        module_context: str | None = None,
    ) -> dict[str, t.Any]: ...
    def broadcast_change(self, var_name: str, value: t.Any): ...
    def broadcast_changes(self, values: dict[str, t.Any] | None = None, **kwargs): ...
    def table_on_edit(self, state: State, var_name: str, payload: dict[str, t.Any]): ...
    def table_on_add(
        self,
        state: State,
        var_name: str,
        payload: dict[str, t.Any],
        new_row: list[t.Any] | None = None,
    ): ...
    def table_on_delete(
        self, state: State, var_name: str, payload: dict[str, t.Any]
    ): ...
    def add_page(self, name: str, page: str | Page, style: str | None = "") -> None: ...
    def add_pages(
        self, pages: t.Mapping[str, str | Page] | str | None = None
    ) -> None: ...
    def add_partial(self, page: str | Page) -> Partial: ...
    def load_config(self, config: Config) -> None: ...
    def get_flask_app(self) -> Flask: ...
    state: Incomplete
    def run(
        self,
        allow_unsafe_werkzeug: bool = ...,
        async_mode: str = ...,
        change_delay: t.Optional[int] = ...,
        chart_dark_template: t.Optional[t.Dict[str, t.Any]] = ...,
        base_url: t.Optional[str] = ...,
        client_url: t.Optional[str] = ...,
        dark_mode: bool = ...,
        dark_theme: t.Optional[t.Dict[str, t.Any]] = ...,
        data_url_max_size: t.Optional[int] = ...,
        debug: bool = ...,
        extended_status: bool = ...,
        favicon: t.Optional[str] = ...,
        flask_log: bool = ...,
        host: str = ...,
        light_theme: t.Optional[t.Dict[str, t.Any]] = ...,
        margin: t.Optional[str] = ...,
        ngrok_token: str = ...,
        notebook_proxy: bool = ...,
        notification_duration: int = ...,
        port: t.Union[t.Literal["auto"], int] = ...,
        port_auto_ranges: t.List[t.Union[int, t.Tuple[int, int]]] = ...,
        propagate: bool = ...,
        run_browser: bool = ...,
        run_in_thread: bool = ...,
        run_server: bool = ...,
        server_config: t.Optional[ServerConfig] = ...,
        single_client: bool = ...,
        state_retention_period: int = ...,
        stylekit: t.Union[bool, Stylekit] = ...,
        system_notification: bool = ...,
        theme: t.Optional[t.Dict[str, t.Any]] = ...,
        time_zone: t.Optional[str] = ...,
        title: t.Optional[str] = ...,
        upload_folder: t.Optional[str] = ...,
        use_arrow: bool = ...,
        use_reloader: bool = ...,
        watermark: t.Optional[str] = ...,
        webapp_path: t.Optional[str] = ...,
    ) -> Flask | None: ...
    def reload(self) -> None: ...
    def stop(self) -> None: ...
    def set_favicon(self, favicon_path: str | Path, state: State | None = None): ...



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import threading
import typing as t

from ._warnings import _warn
from .gui import Gui
from .state import State
from .utils.callable import _is_function


def download(
    state: State, content: t.Any, name: t.Optional[str] = "", on_action: t.Optional[t.Union[str, t.Callable]] = ""
):
    """Download content to the client.

    Arguments:
        state (State^): The current user state as received in any callback.
        content: File path or file content. See below.
        name: File name for the content on the client browser (defaults to content name).
        on_action: Callback function (or callback name) to call when the download ends. See below.

    <h4>Notes:</h4>

    - *content*: this parameter can hold several values depending on your use case:
        - a string: the value must be an existing path name to the file that gets downloaded or
          the URL to the resource you want to download.
        - a buffer (such as a `bytes` object): if the size of the buffer is smaller than the
          [*data_url_max_size*](../../../../../userman/advanced_features/configuration/gui-config.md#p-data_url_max_size)
          configuration setting, then the [`python-magic`](https://pypi.org/project/python-magic/)
          package is used to determine the [MIME type](https://en.wikipedia.org/wiki/Media_type)
          of the buffer content, and the download is performed using a generated "data:" URL with
          the relevant type, and a base64-encoded version of the buffer content.<br/>
          If the buffer is too large, its content is transferred after saving it in a temporary
          server file.
        - *on_action*: this callback is triggered when the transfer of the content is achieved.</br>
          In this function, you can perform any clean-up operation that could be required after
          the download is completed.<br/>
          This callback can use three optional parameters:
            - *state*: the `State^` instance of the caller.
            - *id* (optional): a string representing the identifier of the caller. If this function
              is called directly, this will always be "Gui.download". Some controls may also trigger
              download actions, and then *id* would reflect the identifier of those controls.
            - *payload* (optional): an optional payload from the caller.<br/>
              This is a dictionary with the following keys:
                - *action*: the name of the callback;
                - *args*: an array of two strings. The first element reflects the *name* parameter,
                  and the second element reflects the server-side URL where the file is located.
    """
    if state and isinstance(state._gui, Gui):
        state._invoke_on_gui(Gui._download, content, name, on_action)
    else:
        _warn("'download()' must be called in the context of a callback.")


def notify(
    state: State,
    notification_type: str = "I",
    message: str = "",
    system_notification: t.Optional[bool] = None,
    duration: t.Optional[int] = None,
):
    """Send a notification to the user interface.

    Arguments:
        state (State^): The current user state as received in any callback.
        notification_type: The notification type. This can be one of "success", "info",
            "warning", or "error".<br/>
            To remove the last notification, set this parameter to the empty string.
        message: The text message to display.
        system_notification: If True, the system will also show the notification.<br/>
            If not specified or set to None, this parameter will use the value of
            *configuration[system_notification]*.
        duration: The time, in milliseconds, during which the notification is shown.
            If not specified or set to None, this parameter will use the value of
            *configuration[notification_duration]*.

    Note that you can also call this function with *notification_type* set to the first letter
    or the alert type (i.e. setting *notification_type* to "i" is equivalent to setting it to
    "info").

    If *system_notification* is set to True, then the browser requests the system
    to display a notification as well. They usually appear in small windows that
    fly out of the system tray.<br/>
    The first time your browser is requested to show such a system notification for
    Taipy applications, you may be prompted to authorize the browser to do so. Please
    refer to your browser documentation for details on how to allow or prevent this
    feature.
    """
    if state and isinstance(state._gui, Gui):
        return state._invoke_on_gui(Gui._notify, notification_type, message, system_notification, duration)
    else:
        _warn("'notify()' must be called in the context of a callback.")
        return None


def hold_control(
    state: State,
    callback: t.Optional[t.Union[str, t.Callable]] = None,
    message: t.Optional[str] = "Work in Progress...",
):
    """Hold the User Interface actions.

    When the User Interface is held, users cannot interact with visual elements.<br/>
    The application must call `resume_control()^` so that users can interact again
    with the visual elements.

    You can set a callback function (or the name of a function) in the *callback* parameter. Then,
    a "Cancel" button will be displayed so the user can cancel whatever is happening in the
    application. When pressed, the callback is invoked.

    Arguments:
        state (State^): The current user state received in any callback.
        callback (Optional[Union[str, Callable]]): The function to be called if the user
            chooses to cancel.<br/>
            If empty or None, no cancel action is provided to the user.<br/>
            The signature of this function is:

            - state (`State^`): The user state;
            - id (str): the id of the button that triggered the callback. That will always be
              "UIBlocker" since it is created and managed internally;

            If this parameter is None, no "Cancel" button is displayed.
        message: The message to show. The default value is the string "Work in Progress...".
    """
    if state and isinstance(state._gui, Gui):
        state._invoke_on_gui(Gui._hold_actions, callback, message)
    else:
        _warn("'hold_actions()' must be called in the context of a callback.")


def resume_control(state: State):
    """Resume the User Interface actions.

    This function must be called after `hold_control()^` was invoked, when interaction
    must be allowed again for the user.

    Arguments:
        state (State^): The current user state as received in any callback.
    """
    if state and isinstance(state._gui, Gui):
        state._invoke_on_gui(Gui._resume_actions)
    else:
        _warn("'resume_actions()' must be called in the context of a callback.")


def navigate(
    state: State,
    to: t.Optional[str] = "",
    params: t.Optional[t.Dict[str, str]] = None,
    tab: t.Optional[str] = None,
    force: t.Optional[bool] = False,
):
    """Navigate to a page.

    Arguments:
        state (State^): The current user state as received in any callback.
        to: The name of the page to navigate to. This can be a page identifier (as created by
            `Gui.add_page()^` with no leading '/') or a URL.<br/>
            If omitted, the application navigates to the root page.
        params: A dictionary of query parameters.
        tab: When navigating to a page that is not a known page, the page is opened in a tab identified by
            *tab* (as in [window.open](https://developer.mozilla.org/en-US/docs/Web/API/Window/open)).<br/>
            The default value creates a new tab for the page (which is equivalent to setting *tab* to "_blank").
        force: When navigating to a known page, the content is refreshed even it the page is already shown.
    """
    if state and isinstance(state._gui, Gui):
        state._invoke_on_gui(Gui._navigate, to, params, tab, force)
    else:
        _warn("'navigate()' must be called in the context of a callback.")


def get_user_content_url(
    state: State, path: t.Optional[str] = None, params: t.Optional[t.Dict[str, str]] = None
) -> t.Optional[str]:
    """Get a user content URL.

    This function can be used if you need to deliver dynamic content to your page: you can create
    a path at run-time that, when queried, will deliver some user-defined content defined in the
    *on_user_content* callback (see the description of the `Gui^` class for more information).

    Arguments:
        state (State^): The current user state as received in any callback.
        path: An optional additional path to the URL.
        params: An optional dictionary sent to the *on_user_content* callback.<br/>
            These arguments are added as query parameters to the generated URL and converted into
            strings.

    Returns:
        An URL that, when queried, triggers the *on_user_content* callback.
    """
    if state and isinstance(state._gui, Gui):
        return state._invoke_on_gui(Gui._get_user_content_url, path, params)
    _warn("'get_user_content_url()' must be called in the context of a callback.")
    return None


def get_state_id(state: State) -> t.Optional[str]:
    """Get the identifier of a state.

    The state identifier is a string generated by Taipy GUI for a given `State^` that is used
    to serialize callbacks.
    See the
    [User Manual section on Long Running Callbacks](../../../../../userman/gui/callbacks.md#long-running-callbacks)
    for details on when and how this function can be used.

    Arguments:
        state (State^): The current user state as received in any callback.

    Returns:
        A string that uniquely identifies the state. If this value None, it indicates that *state* is not
        handled by a `Gui^` instance.
    """
    if state and isinstance(state._gui, Gui):
        return state._invoke_on_gui(Gui._get_client_id)
    return None


def get_module_context(state: State) -> t.Optional[str]:
    """Get the name of the module currently in used when using page scopes

    Arguments:
        state (State^): The current user state as received in any callback.

    Returns:
        The name of the current module
    """
    if state and isinstance(state._gui, Gui):
        return state._invoke_on_gui(Gui._get_locals_context)
    return None


def get_context_id(state: State) -> t.Any:
    """NOT DOCUMENTED"""
    _warn("'get_context_id()' was deprecated in Taipy GUI 2.0. Use 'get_state_id()' instead.")
    return get_state_id(state)


def get_module_name_from_state(state: State) -> t.Optional[str]:
    """Get the module name that triggered a callback.

    Pages can be defined in different modules yet refer to callback functions declared elsewhere
    (typically, the application's main module).

    This function returns the name of the module where the page that holds the control that
    triggered the callback was declared.  This lets applications implement different behaviors
    depending on what page is involved.

    This function must be called only in the body of a callback function.

    Arguments:
        state (State^): The `State^` instance, as received in any callback.

    Returns:
        The name of the module that holds the definition of the page containing the control
            that triggered the callback that was provided the *state* object.
    """
    if state and isinstance(state._gui, Gui):
        return state._invoke_on_gui(Gui._get_locals_context)
    return None


def invoke_callback(
    gui: Gui,
    state_id: str,
    callback: t.Callable,
    args: t.Optional[t.Sequence[t.Any]] = None,
    module_context: t.Optional[str] = None,
) -> t.Any:
    """Invoke a user callback for a given state.

    Calling this function is equivalent to calling
    *gui*.`(Gui.)invoke_callback(state_id, callback, args, module_context)^`.

    Arguments:
        gui (Gui^): The current Gui instance.
        state_id: The identifier of the state to use, as returned by `get_state_id()^`.
        callback (Callable[[State^, ...], None]): The user-defined function that is invoked.<br/>
            The first parameter of this function **must** be a `State^`.
        args (Optional[Sequence]): The remaining arguments, as a List or a Tuple.
        module_context (Optional[str]): the name of the module that will be used.
    """
    if isinstance(gui, Gui):
        return gui.invoke_callback(state_id, callback, args, module_context)
    _warn("'invoke_callback()' must be called with a valid Gui instance.")


def broadcast_callback(
    gui: Gui,
    callback: t.Callable,
    args: t.Optional[t.Sequence[t.Any]] = None,
    module_context: t.Optional[str] = None,
) -> t.Dict[str, t.Any]:
    """Invoke a callback for every client.

    Calling this function is equivalent to calling the method
    *gui*.`(Gui.)broadcast_callback(callback, args)^`.

    Arguments:
        gui (Gui^): The current Gui instance.
        callback: The user-defined function to be invoked.<br/>
            The first parameter of this function must be a `State^` object representing the
            client for which it is invoked.<br/>
            The other parameters should reflect the ones provided in the *args* collection.
        args: The parameters to send to *callback*, if any.
    """
    if isinstance(gui, Gui):
        return gui.broadcast_callback(callback, args, module_context)
    _warn("'broadcast_callback()' must be called with a valid Gui instance.")


def invoke_state_callback(gui: Gui, state_id: str, callback: t.Callable, args: t.Union[t.Tuple, t.List]) -> t.Any:
    """NOT DOCUMENTED"""
    _warn("'invoke_state_callback()' was deprecated in Taipy GUI 2.0. Use 'Gui.invoke_callback()' instead.")
    return gui.invoke_callback(state_id, callback, args)


def invoke_long_callback(
    state: State,
    user_function: t.Callable,
    user_function_args: t.Optional[t.Union[t.Tuple, t.List]] = None,
    user_status_function: t.Optional[t.Callable] = None,
    user_status_function_args: t.Optional[t.Union[t.Tuple, t.List]] = None,
    period=0,
):
    """Invoke a long-running user callback.

    Long-running callbacks are run in a separate thread to not block the application itself.

    This function expects to be provided a function to run in the background (in *user_function*).<br/>
    It can also be specified a *status function* that is called when the operation performed by
    *user_function* is finished (successfully or not), or periodically (using the *period* parameter).

    See the
    [User Manual section on Long Running Callbacks](../../../../../userman/gui/callbacks.md#long-running-callbacks)
    for details on when and how this function can be used.

    Arguments:
        state (State^): The `State^` instance, as received in any callback.
        user_function (Callable[[...], None]): The function that will be run independently of Taipy GUI. Note
            that this function must not use *state*, which is not persisted across threads.
        user_function_args (Optional[List|Tuple]): The arguments to send to *user_function*.
        user_status_function (Optional(Callable[[State^, bool, ...], None])): The optional user-defined status
            function that is invoked at the end of and possibly during the runtime of *user_function*:

            - The first parameter of this function is set to a `State^` instance.
            - The second parameter of this function is set to a bool or an int, depending on the
              conditions under which it is called:

               - If this parameter is set to a bool value, then:

                   - If True, this indicates that *user_function* has finished properly.
                        The last argument passed will be the result of the user_function call.
                   - If False, this indicates that *user_function* failed.

               - If this parameter is set to an int value, then this value indicates
                 how many periods (as lengthy as indicated in *period*) have elapsed since *user_function* was
                 started.
        user_status_function_args (Optional[List|Tuple]): The remaining arguments of the user status function.
        period (int): The interval, in milliseconds, at which *user_status_function* is called.<br/>
            The default value is 0, meaning no call to *user_status_function* will happen until *user_function*
            terminates (then the second parameter of that call will be ).</br>
            When set to a value smaller than 500, *user_status_function* is called only when *user_function*
            terminates (as if *period* was set to 0).
    """
    if not state or not isinstance(state._gui, Gui):
        _warn("'invoke_long_callback()' must be called in the context of a callback.")
        return

    if user_status_function_args is None:
        user_status_function_args = []
    if user_function_args is None:
        user_function_args = []

    this_gui = state.get_gui()

    state_id = this_gui._get_client_id()
    module_context = this_gui._get_locals_context()
    if not isinstance(state_id, str) or not isinstance(module_context, str):
        return

    def callback_on_exception(state: State, function_name: str, e: Exception):
        if not this_gui._call_on_exception(function_name, e):
            _warn(f"invoke_long_callback(): Exception raised in function {function_name}()", e)

    def callback_on_status(
        status: t.Union[int, bool],
        e: t.Optional[Exception] = None,
        function_name: t.Optional[str] = None,
        function_result: t.Optional[t.Any] = None,
    ):
        if _is_function(user_status_function):
            this_gui.invoke_callback(
                str(state_id),
                t.cast(t.Callable, user_status_function),
                [status] + list(user_status_function_args) + [function_result],  # type: ignore
                str(module_context),
            )
        if e:
            this_gui.invoke_callback(
                str(state_id),
                callback_on_exception,
                (
                    str(function_name),
                    e,
                ),
                str(module_context),
            )

    def user_function_in_thread(*uf_args):
        try:
            res = user_function(*uf_args)
            callback_on_status(True, function_result=res)
        except Exception as e:
            callback_on_status(False, e, user_function.__name__)

    def thread_status(name: str, period_s: float, count: int):
        active_thread = next((t for t in threading.enumerate() if t.name == name), None)
        if active_thread:
            callback_on_status(count)
            threading.Timer(period_s, thread_status, (name, period_s, count + 1)).start()

    thread = threading.Thread(target=user_function_in_thread, args=user_function_args)
    thread.start()
    if isinstance(period, int) and period >= 500 and _is_function(user_status_function):
        thread_status(thread.name, period / 1000.0, 0)



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import typing as t
from inspect import isclass


def _is_function(s: t.Any) -> bool:
    return callable(s) and not isclass(s)


def _function_name(s: t.Any) -> str:
    if hasattr(s, "__name__"):
        return s.__name__
    elif callable(s):
        return f"<instance of {type(s).__name__}>"
    else:
        return str(s)


def _is_unnamed_function(s: t.Any):
    return (
        (hasattr(s, "__name__") and s.__name__ == "<lambda>")
        or (callable(s) and not hasattr(s, "__name__"))
        or (hasattr(s, "__qualname__") and "<locals>" in s.__qualname__)
    )



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import inspect
import typing as t
from abc import ABCMeta, abstractmethod
from contextlib import nullcontext
from operator import attrgetter
from pathlib import Path
from types import FrameType, SimpleNamespace

from flask import has_app_context

from .utils import _get_module_name_from_frame, _is_in_notebook
from .utils._attributes import _attrsetter

if t.TYPE_CHECKING:
    from .gui import Gui


class State(SimpleNamespace, metaclass=ABCMeta):
    """Accessor to the bound variables from callbacks.

    `State` is used when you need to access the value of variables
    bound to visual elements (see [Binding](../../../../../userman/gui/binding.md)).<br/>
    Because each browser connected to the application server may represent and
    modify any variable at any moment as the result of user interaction, each
    connection holds its own set of variables along with their values. We call
    the set of these the application variables the application _state_, as seen
    by a given client.

    Each callback (see [Callbacks](../../../../../userman/gui/callbacks.md)) receives a specific
    instance of the `State` class, where you can find all the variables bound to
    visual elements in your application.

    Note that `State` also is a Python Context Manager: In situations where you
    have several variables to update, it is more clear and more efficient to assign
    the variable values in a `with` construct:
    ```py
    def my_callback(state, ...):
      ...
      with state as s:
        s.var1 = value1
        s.var2 = value2
      ...
    ```

    You cannot set a variable in the context of a lambda function because Python
    prevents any use of the assignment operator.<br/>
    You can, however, use the `assign()` method on the state that the lambda function
    receives, so you can work around this limitation:

    Here is how you could define a button that changes the value of a variable
    directly in a page expressed using Markdown:
    ```
       <|Set variable|button|on_action={lambda s: s.assign("var_name", new_value)}|>
    ```
    This would be strictly similar to the Markdown line:
    ```
       <|Set variable|button|on_action=change_variable|>
    ```
    with the Python code:
    ```py
    def change_variable(state):
        state.var_name = new_value
    ```
    """

    def __init__(self) -> None:
        self._gui: "Gui"

    @abstractmethod
    def get_gui(self) -> "Gui":
        """Return the Gui instance for this state object.

        Returns:
            Gui: The Gui instance for this state object.
        """
        raise NotImplementedError

    def assign(self, name: str, value: t.Any) -> t.Any:
        """Assign a value to a state variable.

        This should be used only from within a lambda function used
        as a callback in a visual element.

        Arguments:
            name (str): The variable name to assign to.
            value (Any): The new variable value.

        Returns:
            Any: The previous value of the variable.
        """
        val = attrgetter(name)(self)
        _attrsetter(self, name, value)
        return val

    def refresh(self, name: str):
        """Refresh a state variable.

        This allows to re-sync the user interface with a variable value.

        Arguments:
            name (str): The variable name to refresh.
        """
        val = attrgetter(name)(self)
        _attrsetter(self, name, val)

    def _set_context(self, gui: "Gui") -> t.ContextManager[None]:
        return nullcontext()

    def broadcast(self, name: str, value: t.Any):
        """Update a variable on all clients.

        All connected clients will receive an update of the variable called *name* with the
        provided value, even if it is not shared.

        Arguments:
            name (str): The variable name to update.
            value (Any): The new variable value.
        """
        with self._set_context(self._gui):
            encoded_name = self._gui._bind_var(name)
            self._gui._broadcast_all_clients(encoded_name, value)

    def __enter__(self):
        self._gui.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        return self._gui.__exit__(exc_type, exc_value, traceback)

    def set_favicon(self, favicon_path: t.Union[str, Path]):
        """Change the favicon for the client of this state.

        This function dynamically changes the favicon (the icon associated with the application's
        pages) of Taipy GUI pages for the specific client of this state.

        Note that the *favicon* parameter to `(Gui.)run()^` can also be used to change
        the favicon when the application starts.

        Arguments:
            favicon_path: The path to the image file to use.<br/>
                This can be expressed as a path name or a URL (relative or not).
        """
        self._gui.set_favicon(favicon_path, self)

    @abstractmethod
    def __getitem__(self, key: str) -> "State": ...

    @abstractmethod
    def _invoke_on_gui(self, method: t.Callable, *args: t.Any) -> t.Any: ...


class _GuiState(State):
    __gui_attr = "_gui"
    __attrs = (
        __gui_attr,
        "_user_var_list",
        "_context_list",
    )
    __methods = (
        "assign",
        "broadcast",
        "get_gui",
        "refresh",
        "set_favicon",
        "_set_context",
        "_notebook_context",
        "_get_placeholder",
        "_set_placeholder",
        "_get_gui_attr",
        "_get_placeholder_attrs",
        "_add_attribute",
        "_invoke_on_gui",
    )
    __placeholder_attrs = ("_taipy_p1", "_current_context", "__state_id")
    __excluded_attrs = __attrs + __methods + __placeholder_attrs

    def __init__(self, gui: "Gui", var_list: t.Iterable[str], context_list: t.Iterable[str]) -> None:
        super().__setattr__(
            _GuiState.__attrs[1], list(_GuiState.__filter_var_list(var_list, _GuiState.__excluded_attrs))
        )
        super().__setattr__(_GuiState.__attrs[2], list(context_list))
        super().__setattr__(_GuiState.__attrs[0], gui)
        super().__init__()

    @staticmethod
    def __filter_var_list(var_list: t.Iterable[str], excluded_attrs: t.Iterable[str]) -> t.Iterable[str]:
        return filter(lambda n: n not in excluded_attrs, var_list)

    def get_gui(self) -> "Gui":
        return super().__getattribute__(_GuiState.__gui_attr)

    def __getattribute__(self, name: str) -> t.Any:
        if name == "__class__":
            return _GuiState
        if name in _GuiState.__methods:
            return super().__getattribute__(name)
        gui: "Gui" = self.get_gui()
        if name == _GuiState.__gui_attr:
            return gui
        if name in _GuiState.__excluded_attrs:
            raise AttributeError(f"Variable '{name}' is protected and is not accessible.")
        if gui._is_in_brdcst_callback() and (
            name not in gui._get_shared_variables() and not gui._bindings()._is_single_client()
        ):
            raise AttributeError(f"Variable '{name}' is not available to be accessed in shared callback.")
        if not name.startswith("__") and name not in super().__getattribute__(_GuiState.__attrs[1]):
            raise AttributeError(f"Variable '{name}' is not defined.")
        with self._notebook_context(gui), self._set_context(gui):
            encoded_name = gui._bind_var(name)
            return getattr(gui._bindings(), encoded_name)

    def __setattr__(self, name: str, value: t.Any) -> None:
        gui: "Gui" = super().__getattribute__(_GuiState.__gui_attr)
        if gui._is_in_brdcst_callback() and (
            name not in gui._get_shared_variables() and not gui._bindings()._is_single_client()
        ):
            raise AttributeError(f"Variable '{name}' is not available to be accessed in shared callback.")
        if not name.startswith("__") and name not in super().__getattribute__(_GuiState.__attrs[1]):
            raise AttributeError(f"Variable '{name}' is not accessible.")
        with self._notebook_context(gui), self._set_context(gui):
            encoded_name = gui._bind_var(name)
            setattr(gui._bindings(), encoded_name, value)

    def __getitem__(self, key: str):
        context = key if key in super().__getattribute__(_GuiState.__attrs[2]) else None
        if context is None:
            gui: "Gui" = super().__getattribute__(_GuiState.__gui_attr)
            page_ctx = gui._get_page_context(key)
            context = page_ctx if page_ctx is not None else None
        if context is None:
            raise RuntimeError(f"Can't resolve context '{key}' from state object")
        self._set_placeholder(_GuiState.__placeholder_attrs[1], context)
        return self

    def _set_context(self, gui: "Gui") -> t.ContextManager[None]:
        if (pl_ctx := self._get_placeholder(_GuiState.__placeholder_attrs[1])) is not None:
            self._set_placeholder(_GuiState.__placeholder_attrs[1], None)
            if pl_ctx != gui._get_locals_context():
                return gui._set_locals_context(pl_ctx)
        if len(inspect.stack()) > 1:
            ctx = _get_module_name_from_frame(t.cast(FrameType, t.cast(FrameType, inspect.stack()[2].frame)))
            current_context = gui._get_locals_context()
            # ignore context if the current one starts with the new one (to resolve for class modules)
            if ctx != current_context and not current_context.startswith(str(ctx)):
                return gui._set_locals_context(ctx)
        return nullcontext()

    def _notebook_context(self, gui: "Gui"):
        return gui.get_flask_app().app_context() if not has_app_context() and _is_in_notebook() else nullcontext()

    def _get_placeholder(self, name: str):
        if name in _GuiState.__placeholder_attrs:
            try:
                return super().__getattribute__(name)
            except AttributeError:
                return None
        return None

    def _set_placeholder(self, name: str, value: t.Any):
        if name in _GuiState.__placeholder_attrs:
            super().__setattr__(name, value)

    def _get_placeholder_attrs(self):
        return _GuiState.__placeholder_attrs

    def _add_attribute(self, name: str, default_value: t.Optional[t.Any] = None) -> bool:
        attrs: t.List[str] = super().__getattribute__(_GuiState.__attrs[1])
        if name not in attrs:
            attrs.append(name)
            gui = super().__getattribute__(_GuiState.__gui_attr)
            return gui._bind_var_val(name, default_value)
        return False

    def _invoke_on_gui(self, method: t.Callable, *args):
        return method(self.get_gui(), *args)


class _AsyncState(_GuiState):
    def __init__(self, state: State) -> None:
        super().__init__(state.get_gui(), [], [])
        self._set_placeholder("__state_id", state.get_gui()._get_client_id())

    @staticmethod
    def __set_var_in_state(state: State, var_name: str, value: t.Any):
        setattr(state, var_name, value)

    @staticmethod
    def __get_var_from_state(state: State, var_name: str):
        return getattr(state, var_name)

    def __setattr__(self, var_name: str, var_value: t.Any) -> None:
        self.get_gui().invoke_callback(
            t.cast(str, self._get_placeholder("__state_id")), _AsyncState.__set_var_in_state, [var_name, var_value]
        )

    def __getattr__(self, var_name: str) -> t.Any:
        return self.get_gui().invoke_callback(
            t.cast(str, self._get_placeholder("__state_id")), _AsyncState.__get_var_from_state, [var_name]
        )

    def _invoke_on_gui(self, method: t.Callable, *args):
        return self.get_gui()._invoke_method(t.cast(str, self._get_placeholder("__state_id")), method, *args)  # type: ignore[reportAttributeAccessIssue]



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import inspect
import json
import os
import sys
import types
import typing as t

from taipy.common.logger._taipy_logger import _TaipyLogger

from ..utils.singleton import _Singleton
from ..utils.viselements import VisElementProperties, VisElements, resolve_inherits
from ._element import _Block, _Control

if t.TYPE_CHECKING:
    from ..extension.library import ElementLibrary


class _ElementApiGenerator(object, metaclass=_Singleton):
    def __init__(self) -> None:
        self.__module: t.Optional[types.ModuleType] = None

    @staticmethod
    def find_default_property(property_list: t.List[VisElementProperties]) -> str:
        for property in property_list:
            if property.get("default_property", False) is True:
                return property["name"]
        return ""

    @staticmethod
    def get_properties_dict(property_list: t.List[VisElementProperties]) -> t.Dict[str, t.Any]:
        return {prop["name"]: prop.get("type", "str") for prop in property_list}

    def add_default(self):
        if self.__module is not None:
            return
        current_frame = inspect.currentframe()
        error_message = "Cannot generate elements API for the current module"
        if current_frame is None:
            raise RuntimeError(f"{error_message}: No frame found.")
        if current_frame.f_back is None:
            raise RuntimeError(f"{error_message}: taipy-gui module not found.")
        module_name = current_frame.f_back.f_globals["__name__"]
        self.__module = module = sys.modules[module_name]
        with open(os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "viselements.json"))) as f:
            viselements_json = json.load(f)
            if "blocks" not in viselements_json or "controls" not in viselements_json:
                raise RuntimeError(f"{error_message}: Invalid viselements.json file.")
            viselements = resolve_inherits(t.cast(VisElements, viselements_json))
            for blockElement in viselements["blocks"]:
                default_property = _ElementApiGenerator.find_default_property(blockElement[1]["properties"])
                setattr(
                    module,
                    blockElement[0],
                    _ElementApiGenerator.create_block_api(
                        blockElement[0],
                        blockElement[0],
                        default_property,
                        _ElementApiGenerator.get_properties_dict(blockElement[1]["properties"]),
                    ),
                )
            for controlElement in viselements["controls"]:
                default_property = _ElementApiGenerator.find_default_property(controlElement[1]["properties"])
                setattr(
                    module,
                    controlElement[0],
                    _ElementApiGenerator.create_control_api(
                        controlElement[0],
                        controlElement[0],
                        default_property,
                        _ElementApiGenerator.get_properties_dict(controlElement[1]["properties"]),
                    ),
                )

    def add_library(self, library: "ElementLibrary"):
        library_name = library.get_name()
        if self.__module is None:
            _TaipyLogger._get_logger().info(
                f"Python API for extension library '{library_name}' is not available. To fix this, import 'taipy.gui.builder' before importing the extension library."  # noqa: E501
            )
            return
        library_module = getattr(self.__module, library_name, None)
        if library_module is None:
            library_module = types.ModuleType(library_name)
            setattr(self.__module, library_name, library_module)
        for element_name, element in library.get_elements().items():
            setattr(
                library_module,
                element_name,
                _ElementApiGenerator().create_control_api(
                    element_name,
                    f"{library_name}.{element_name}",
                    element.default_attribute,
                    {name: str(prop.property_type) for name, prop in element.attributes.items()},
                ),
            )
            # Allow element to be accessed from the root module
            if hasattr(self.__module, element_name):
                _TaipyLogger._get_logger().info(
                    f"Can't add element `{element_name}` of library `{library_name}` to the root of Builder API as another element with the same name already exists."  # noqa: E501
                )
                continue
            setattr(self.__module, element_name, getattr(library_module, element_name))

    @staticmethod
    def create_block_api(
        classname: str,
        element_name: str,
        default_property: str,
        properties: t.Dict[str, str],
    ):
        return _ElementApiGenerator.create_element_api(classname, element_name, default_property, properties, _Block)

    @staticmethod
    def create_control_api(
        classname: str,
        element_name: str,
        default_property: str,
        properties: t.Dict[str, str],
    ):
        return _ElementApiGenerator.create_element_api(classname, element_name, default_property, properties, _Control)

    @staticmethod
    def create_element_api(
        classname: str,
        element_name: str,
        default_property: str,
        properties: t.Dict[str, str],
        ElementBaseClass: t.Union[t.Type[_Block], t.Type[_Control]],
    ):
        return type(
            classname,
            (ElementBaseClass,),
            {
                "_ELEMENT_NAME": element_name,
                "_DEFAULT_PROPERTY": default_property,
                "_TYPES": {f"{parts[0]}__" if len(parts := k.split("[")) > 1 else k: v for k, v in properties.items()},
            },
        )



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import typing as t
from types import LambdaType

from ._variable_directory import _variable_encode


def _get_lambda_id(lambda_fn: LambdaType, module: t.Optional[str] = None, index: t.Optional[int] = None):
    return _variable_encode(
        f"__lambda_{id(lambda_fn)}{f'_{index}' if index is not None else ''}", lambda_fn.__module__ or module
    )



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.


import re
import typing as t
from enum import Enum

from .._renderers.utils import _get_columns_dict
from .._warnings import _warn
from ..types import PropertyType
from ..utils import _MapDict

if t.TYPE_CHECKING:
    from ..gui import Gui


class _Chart_iprops(Enum):
    x = 0
    y = 1
    z = 2
    label = 3
    text = 4
    mode = 5
    type = 6
    color = 7
    xaxis = 8
    yaxis = 9
    selected_color = 10
    marker = 11
    selected_marker = 12
    orientation = 13
    _name = 14
    line = 15
    text_anchor = 16
    options = 17
    lon = 18
    lat = 19
    base = 20
    r = 21
    theta = 22
    close = 23
    open = 24
    high = 25
    low = 26
    locations = 27
    values = 28
    labels = 29
    decimator = 30
    measure = 31
    parents = 32


__CHART_AXIS: t.Dict[str, t.Tuple[_Chart_iprops, ...]] = {
    "bar": (_Chart_iprops.x, _Chart_iprops.y, _Chart_iprops.base),
    "candlestick": (
        _Chart_iprops.x,
        _Chart_iprops.close,
        _Chart_iprops.open,
        _Chart_iprops.high,
        _Chart_iprops.low,
    ),
    "choropleth": (_Chart_iprops.locations, _Chart_iprops.z),
    "choroplethmap": (_Chart_iprops.locations, _Chart_iprops.z),
    "choroplethmapbox": (_Chart_iprops.locations, _Chart_iprops.z),
    "densitymap": (_Chart_iprops.lon, _Chart_iprops.lat, _Chart_iprops.z),
    "densitymapbox": (_Chart_iprops.lon, _Chart_iprops.lat, _Chart_iprops.z),
    "funnelarea": (_Chart_iprops.values,),
    "pie": (_Chart_iprops.values, _Chart_iprops.labels),
    "scattergeo": (_Chart_iprops.lon, _Chart_iprops.lat),
    "scattermap": (_Chart_iprops.lon, _Chart_iprops.lat),
    "scattermapbox": (_Chart_iprops.lon, _Chart_iprops.lat),
    "scatterpolar": (_Chart_iprops.r, _Chart_iprops.theta),
    "scatterpolargl": (_Chart_iprops.r, _Chart_iprops.theta),
    "treemap": (_Chart_iprops.labels, _Chart_iprops.parents, _Chart_iprops.values),
    "waterfall": (_Chart_iprops.x, _Chart_iprops.y, _Chart_iprops.measure),
}
__CHART_DEFAULT_AXIS: t.Tuple[_Chart_iprops, ...] = (_Chart_iprops.x, _Chart_iprops.y, _Chart_iprops.z)
__CHART_MARKER_TO_COLS: t.Tuple[str, ...] = ("color", "size", "symbol", "opacity", "colors")
__CHART_NO_INDEX: t.Tuple[str, ...] = ("pie", "histogram", "heatmap", "funnelarea")
_CHART_NAMES: t.Tuple[str, ...] = tuple(e.name[1:] if e.name[0] == "_" else e.name for e in _Chart_iprops)


def __check_dict(values: t.List[t.Any], properties: t.Iterable[_Chart_iprops]) -> None:
    for prop in properties:
        if values[prop.value] is not None and not isinstance(values[prop.value], (dict, _MapDict)):
            _warn(f"Property {prop.name} of chart control should be a dict.")
            values[prop.value] = None


def __get_multiple_indexed_attributes(
    attributes: t.Dict[str, t.Any], names: t.Iterable[str], index: t.Optional[int] = None
) -> t.List[t.Optional[str]]:
    names = names if index is None else [f"{n}[{index}]" for n in names]  # type: ignore
    return [attributes.get(name) for name in names]


__RE_INDEXED_DATA = re.compile(r"^(\d+)\/(.*)")


def __get_col_from_indexed(col_name: str, idx: int) -> t.Optional[str]:
    if re_res := __RE_INDEXED_DATA.search(col_name):
        return col_name if str(idx) == re_res.group(1) else None
    return col_name


def _build_chart_config(gui: "Gui", attributes: t.Dict[str, t.Any], col_types: t.Dict[str, str]):  # noqa: C901
    if ("data" not in attributes and "figure" in attributes) or (attributes.get("data") is None):
        return {"traces": []}
    default_type = attributes.get("_default_type", "scatter")
    default_mode = attributes.get("_default_mode", "lines+markers")
    trace = __get_multiple_indexed_attributes(attributes, _CHART_NAMES)
    if not trace[_Chart_iprops.mode.value]:
        trace[_Chart_iprops.mode.value] = default_mode
    # type
    if not trace[_Chart_iprops.type.value]:
        trace[_Chart_iprops.type.value] = default_type
    if not trace[_Chart_iprops.xaxis.value]:
        trace[_Chart_iprops.xaxis.value] = "x"
    if not trace[_Chart_iprops.yaxis.value]:
        trace[_Chart_iprops.yaxis.value] = "y"
    # Indexed properties: Check for arrays
    for prop in _Chart_iprops:
        values = trace[prop.value]
        if isinstance(values, (list, tuple)) and len(values):
            prop_name = prop.name[1:] if prop.name[0] == "_" else prop.name
            for idx, val in enumerate(values):
                if idx == 0:
                    trace[prop.value] = val
                if val is not None:
                    indexed_prop = f"{prop_name}[{idx + 1}]"
                    if attributes.get(indexed_prop) is None:
                        attributes[indexed_prop] = val
    # marker selected_marker options
    __check_dict(trace, (_Chart_iprops.marker, _Chart_iprops.selected_marker, _Chart_iprops.options))
    axis = []
    traces: t.List[t.List[t.Optional[str]]] = []
    idx = 1
    indexed_trace = __get_multiple_indexed_attributes(attributes, _CHART_NAMES, idx)
    if len([x for x in indexed_trace if x]):
        while len([x for x in indexed_trace if x]):
            axis.append(
                __CHART_AXIS.get(
                    indexed_trace[_Chart_iprops.type.value] or trace[_Chart_iprops.type.value] or "",
                    __CHART_DEFAULT_AXIS,
                )
            )
            # marker selected_marker options
            __check_dict(indexed_trace, (_Chart_iprops.marker, _Chart_iprops.selected_marker, _Chart_iprops.options))
            if trace[_Chart_iprops.decimator.value] and not indexed_trace[_Chart_iprops.decimator.value]:
                # copy the decimator only once
                indexed_trace[_Chart_iprops.decimator.value] = trace[_Chart_iprops.decimator.value]
                trace[_Chart_iprops.decimator.value] = None
            traces.append([x or trace[i] for i, x in enumerate(indexed_trace)])
            idx += 1
            indexed_trace = __get_multiple_indexed_attributes(attributes, _CHART_NAMES, idx)
    else:
        traces.append(trace)
        # axis names
        axis.append(__CHART_AXIS.get(trace[_Chart_iprops.type.value] or "", __CHART_DEFAULT_AXIS))

    # list of data columns name indexes with label text
    dt_idx = tuple(e.value for e in (axis[0] + (_Chart_iprops.label, _Chart_iprops.text)))

    # configure columns
    columns: t.Set[str] = set()
    for j, trace in enumerate(traces):
        dt_idx = tuple(
            e.value for e in (axis[j] if j < len(axis) else axis[0]) + (_Chart_iprops.label, _Chart_iprops.text)
        )
        columns.update([trace[i] or "" for i in dt_idx if trace[i]])
    # add optional column if any
    markers = [
        t[_Chart_iprops.marker.value]
        or ({"color": t[_Chart_iprops.color.value]} if t[_Chart_iprops.color.value] else None)
        for t in traces
    ]
    opt_cols = set()
    for m in markers:
        if isinstance(m, (dict, _MapDict)):
            for prop1 in __CHART_MARKER_TO_COLS:
                val = m.get(prop1)
                if isinstance(val, str) and val not in columns:
                    opt_cols.add(val)

    # Validate the column names
    col_dict = _get_columns_dict(attributes.get("data"), list(columns), col_types, opt_columns=opt_cols)

    # Manage Decimator
    decimators: t.List[t.Optional[str]] = []
    for tr in traces:
        if tr[_Chart_iprops.decimator.value]:
            cls = gui._get_user_instance(
                class_name=str(tr[_Chart_iprops.decimator.value]), class_type=PropertyType.decimator.value
            )
            if isinstance(cls, PropertyType.decimator.value):
                decimators.append(str(tr[_Chart_iprops.decimator.value]))
                continue
        decimators.append(None)

    # set default columns if not defined
    icols = [
        [c2 for c2 in [__get_col_from_indexed(c1, i) for c1 in t.cast(dict, col_dict).keys()] if c2]
        for i in range(len(traces))
    ]

    for i, tr in enumerate(traces):
        if i < len(axis):
            used_cols = {tr[ax.value] for ax in axis[i] if tr[ax.value]}
            unused_cols = [c for c in icols[i] if c not in used_cols]
            if unused_cols and not any(tr[ax.value] for ax in axis[i]):
                traces[i] = [
                    v or (unused_cols.pop(0) if unused_cols and _Chart_iprops(j) in axis[i] else v)
                    for j, v in enumerate(tr)
                ]

    if col_dict is not None:
        reverse_cols = {str(cd.get("dfid")): c for c, cd in col_dict.items()}

        # List used axis
        used_axis = [[e for e in (axis[j] if j < len(axis) else axis[0]) if tr[e.value]] for j, tr in enumerate(traces)]

        ret_dict = {
            "columns": col_dict,
            "labels": [
                reverse_cols.get(tr[_Chart_iprops.label.value] or "", (tr[_Chart_iprops.label.value] or ""))
                for tr in traces
            ],
            "texts": [
                reverse_cols.get(tr[_Chart_iprops.text.value] or "", (tr[_Chart_iprops.text.value] or None))
                for tr in traces
            ],
            "modes": [tr[_Chart_iprops.mode.value] for tr in traces],
            "types": [tr[_Chart_iprops.type.value] for tr in traces],
            "xaxis": [tr[_Chart_iprops.xaxis.value] for tr in traces],
            "yaxis": [tr[_Chart_iprops.yaxis.value] for tr in traces],
            "markers": markers,
            "selectedMarkers": [
                tr[_Chart_iprops.selected_marker.value]
                or (
                    {"color": tr[_Chart_iprops.selected_color.value]}
                    if tr[_Chart_iprops.selected_color.value]
                    else None
                )
                for tr in traces
            ],
            "traces": [
                [reverse_cols.get(c or "", c) for c in [tr[e.value] for e in used_axis[j]]]
                for j, tr in enumerate(traces)
            ],
            "orientations": [tr[_Chart_iprops.orientation.value] for tr in traces],
            "names": [tr[_Chart_iprops._name.value] for tr in traces],
            "lines": [
                (
                    tr[_Chart_iprops.line.value]
                    if isinstance(tr[_Chart_iprops.line.value], (dict, _MapDict))
                    else {"dash": tr[_Chart_iprops.line.value]}
                    if tr[_Chart_iprops.line.value]
                    else None
                )
                for tr in traces
            ],
            "textAnchors": [tr[_Chart_iprops.text_anchor.value] for tr in traces],
            "options": [tr[_Chart_iprops.options.value] for tr in traces],
            "axisNames": [[e.name for e in ax] for ax in used_axis],
            "addIndex": [tr[_Chart_iprops.type.value] not in __CHART_NO_INDEX for tr in traces],
        }
        if len([d for d in decimators if d]):
            ret_dict.update(decimators=decimators)
        return ret_dict
    return {}



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

import typing as t
from enum import Enum
from inspect import isclass
from operator import add

from .._warnings import _warn
from ..icon import Icon
from . import _MapDict


class _AdaptedLov:
    def __init__(self, lov: t.Any, var_type: str) -> None:
        self._lov = lov
        self._type = var_type

    @staticmethod
    def get_lov(lov: t.Any):
        return lov._lov if isinstance(lov, _AdaptedLov) else lov

    @staticmethod
    def get_type(lov: t.Any):
        return lov._type if isinstance(lov, _AdaptedLov) else None


class _Adapter:
    def __init__(self) -> None:
        self.__adapter_for_type: t.Dict[str, t.Callable] = {}
        self.__type_for_variable: t.Dict[str, str] = {}
        self.__warning_by_type: t.Set[str] = set()

    def _get_adapted_lov(self, lov: t.Any, var_type: str) -> _AdaptedLov:
        if not isinstance(lov, list) and hasattr(lov, "tolist"):
            try:
                lov = lov.tolist()  # type: ignore[union-attr]
            except Exception as e:
                _warn("Error accessing List of values", e)
        return _AdaptedLov(lov, var_type)

    def _add_for_type(self, type_name: str, adapter: t.Callable) -> None:
        self.__adapter_for_type[type_name] = adapter

    def _add_type_for_var(self, var_name: str, type_name: str) -> None:
        self.__type_for_variable[var_name] = type_name

    def _get_for_type(self, type_name: str) -> t.Optional[t.Callable]:
        return self.__adapter_for_type.get(type_name)

    def _get_unique_type(self, type_name: str) -> str:
        index = 0
        while type_name in self.__adapter_for_type:
            type_name = f"{type_name}{index}"
            index += 1
        return type_name

    def run(self, var_name: str, value: t.Any, id_only=False) -> t.Any:
        lov = _AdaptedLov.get_lov(value)
        adapter = self.__get_for_var(var_name, value)
        if isclass(lov) and issubclass(lov, Enum):
            lov = list(lov)
        if isinstance(lov, (list, tuple)):
            res = []
            for elt in lov:
                v = self._run(adapter, elt, var_name, id_only)
                res.append(v if v is not None else elt)
            return res
        return self._run(adapter, lov, var_name, id_only)

    def __get_for_var(self, var_name: str, value: t.Any) -> t.Optional[t.Callable]:
        adapter = None
        type_name = _AdaptedLov.get_type(value)
        if type_name:
            adapter = self.__adapter_for_type.get(type_name)
        if callable(adapter):
            return adapter
        type_name = self.__type_for_variable.get(var_name)
        if not isinstance(type_name, str):
            adapter = self.__adapter_for_type.get(var_name)
            lov = _AdaptedLov.get_lov(value)
            elt = lov[0] if isinstance(lov, (list, tuple)) and len(lov) else None
            type_name = var_name if callable(adapter) else type(elt).__name__
        if adapter is None:
            adapter = self.__adapter_for_type.get(type_name)
        return adapter if callable(adapter) else None

    def _get_elt_per_ids(
        self, var_name: str, lov: t.List[t.Any], adapter: t.Optional[t.Callable] = None
    ) -> t.Dict[str, t.Any]:
        dict_res = {}
        type_name = _AdaptedLov.get_type(lov)
        lov = _AdaptedLov.get_lov(lov)
        if not adapter and type_name:
            adapter = self.__adapter_for_type.get(type_name)
        if not adapter:
            adapter = self.__get_for_var(var_name, lov[0] if lov else None)
        for value in lov:
            try:
                result = adapter(value._dict if isinstance(value, _MapDict) else value) if adapter else value
                if result is not None:
                    dict_res[self.__get_id(result)] = value
                    children = self.__get_children(result)
                    if children is not None:
                        dict_res.update(self._get_elt_per_ids(var_name, children, adapter))
            except Exception as e:
                _warn(f"Cannot run adapter for {var_name}", e)
        return dict_res

    def _run(
        self, adapter: t.Optional[t.Callable], value: t.Any, var_name: str, id_only=False
    ) -> t.Union[t.Tuple[str, ...], str, None]:
        if value is None:
            return None
        try:
            result = value._dict if isinstance(value, _MapDict) else value
            if adapter:
                result = adapter(result)
                if result is None:
                    return result
            elif isinstance(result, (str, float, int)):
                return str(result)
            tpl_res = self._get_valid_result(result, id_only)
            if tpl_res is None:
                _warn(
                    f"Adapter for {var_name} did not return a valid result. Please check the documentation on List of Values Adapters."  # noqa: E501
                )
            else:
                if not id_only and len(tpl_res) > 2 and isinstance(tpl_res[2], list) and len(tpl_res[2]) > 0:
                    tpl_res = (tpl_res[0], tpl_res[1], self.__on_tree(adapter, tpl_res[2]))
                return (
                    add(type(result)(tpl_res), result[len(tpl_res) :])
                    if isinstance(result, (tuple, list)) and isinstance(tpl_res, (tuple, list))
                    else tpl_res # type: ignore[reportReturnType]
                )
        except Exception as e:
            _warn(f"Cannot run adapter for {var_name}", e)
        return None

    def __on_tree(self, adapter: t.Optional[t.Callable], tree: t.List[t.Any]):
        ret_list = []
        for elt in tree:
            ret = self._run(adapter, elt, adapter.__name__ if adapter else "adapter")
            if ret is not None:
                ret_list.append(ret)
        return ret_list

    def _get_valid_result(self, value: t.Any, id_only=False) -> t.Union[t.Tuple[str, ...], str, None]:
        id = self.__get_id(value)
        if id_only:
            return id
        label = self.__get_label(value)
        if label is None:
            return None
        children = self.__get_children(value)
        return (id, label) if children is None else (id, label, children)  # type: ignore

    def __get_id(self, value: t.Any, dig=True) -> str:
        if isinstance(value, str):
            return value
        elif dig:
            if isinstance(value, (list, tuple)) and len(value):
                return self.__get_id(value[0], False)
            elif hasattr(value, "id"):
                return self.__get_id(t.cast(t.Any, value).id, False)
            elif hasattr(value, "__getitem__") and "id" in value:
                return self.__get_id(t.cast(dict, value).get("id"), False)
        if value is not None and type(value).__name__ not in self.__warning_by_type:
            _warn(f"LoV id must be a string, using a string representation of {type(value)}.")
            self.__warning_by_type.add(type(value).__name__)
        return "" if value is None else str(value)

    def __get_label(self, value: t.Any, dig=True) -> t.Union[str, t.Dict, None]:
        if isinstance(value, (str, Icon)):
            return Icon.get_dict_or(value)
        elif dig:
            if isinstance(value, (list, tuple)) and len(value) > 1:
                return self.__get_label(value[1], False)
            elif hasattr(value, "label"):
                return self.__get_label(t.cast(t.Any, value).label, False)
            elif hasattr(value, "__getitem__") and "label" in value:
                return self.__get_label(t.cast(dict, value).get("label"), False)
        return None

    def __get_children(self, value: t.Any) -> t.Optional[t.List[t.Any]]:
        if isinstance(value, (tuple, list)) and len(value) > 2:
            return value[2] if isinstance(value[2], list) else None if value[2] is None else [value[2]]
        elif hasattr(value, "children"):
            return (
                t.cast(t.Any, value).children
                if isinstance(t.cast(t.Any, value).children, list)
                else None
                if t.cast(t.Any, value).children is None
                else [t.cast(t.Any, value).children]
            )
        elif hasattr(value, "__getitem__") and "children" in value:
            return (
                t.cast(dict, value).get("children")
                if isinstance(t.cast(dict, value).get("children"), list)
                else None
                if t.cast(dict, value).get("children") is None
                else [t.cast(dict, value).get("children")]
            )
        return None



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import os
import typing as t
from datetime import datetime
from importlib import util
from tempfile import mkstemp

import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype

from .._warnings import _warn
from ..gui import Gui
from ..types import PropertyType
from ..utils import _RE_PD_TYPE, _get_date_col_str_name
from .comparison import _compare_function
from .data_accessor import _DataAccessor
from .data_format import _DataFormat

_has_arrow_module = False
if util.find_spec("pyarrow"):
    _has_arrow_module = True
    import pyarrow as pa


class _PandasDataAccessor(_DataAccessor):
    __types = (pd.DataFrame, pd.Series)

    __INDEX_COL = "_tp_index"

    __AGGREGATE_FUNCTIONS: t.List[str] = ["count", "sum", "mean", "median", "min", "max", "std", "first", "last"]

    def to_pandas(self, value: t.Union[pd.DataFrame, pd.Series]) -> t.Union[t.List[pd.DataFrame], pd.DataFrame]:
        return self.__to_dataframe(value)

    def __to_dataframe(self, value: t.Union[pd.DataFrame, pd.Series]) -> pd.DataFrame:
        if isinstance(value, pd.Series):
            return pd.DataFrame(value)
        return t.cast(pd.DataFrame, value)

    def _from_pandas(self, value: pd.DataFrame, data_type: t.Type) -> t.Any:
        if data_type is pd.Series:
            return value.iloc[:, 0]
        return value

    @staticmethod
    def get_supported_classes() -> t.List[t.Type]:
        return list(_PandasDataAccessor.__types)

    @staticmethod
    def __user_function(
        row: pd.Series, gui: Gui, column_name: t.Optional[str], user_function: t.Callable, function_name: str
    ) -> str:  # pragma: no cover
        args = []
        if column_name:
            args.append(row[column_name])
        args.extend((row.name, row))  # type: ignore[arg-type]
        if column_name:
            args.append(column_name)  # type: ignore[arg-type]
        try:
            return str(gui._call_function_with_state(user_function, args))
        except Exception as e:
            _warn(f"Exception raised when calling user function {function_name}()", e)
        return ""

    def __is_date_column(self, data: pd.DataFrame, col_name: str) -> bool:
        col_types = data.dtypes[data.dtypes.index.astype(str) == col_name]
        return len(col_types[col_types.astype(str).str.startswith("datetime")]) > 0  # type: ignore

    def __build_transferred_cols(
        self,
        payload_cols: t.Any,
        dataframe: pd.DataFrame,
        styles: t.Optional[t.Dict[str, str]] = None,
        tooltips: t.Optional[t.Dict[str, str]] = None,
        is_copied: t.Optional[bool] = False,
        new_indexes: t.Optional[np.ndarray] = None,
        handle_nan: t.Optional[bool] = False,
        formats: t.Optional[t.Dict[str, str]] = None,
    ) -> pd.DataFrame:
        dataframe = dataframe.iloc[new_indexes] if new_indexes is not None else dataframe
        if isinstance(payload_cols, list) and len(payload_cols):
            col_types = dataframe.dtypes[dataframe.dtypes.index.astype(str).isin(payload_cols)]
        else:
            col_types = dataframe.dtypes
        cols = col_types.index.astype(str).tolist()
        new_cols = {}
        if styles:
            for k, v in styles.items():
                col_applied = ""
                new_data = None
                func = self._gui._get_user_function(v)
                if callable(func):
                    col_applied, new_data = self.__apply_user_function(
                        func, k if k in cols else None, v, dataframe, "tps__"
                    )
                new_cols[col_applied or v] = new_data if col_applied else v
        if tooltips:
            for k, v in tooltips.items():
                func = self._gui._get_user_function(v)
                if callable(func):
                    col_applied, new_data = self.__apply_user_function(
                        func, k if k in cols else None, v, dataframe, "tpt__"
                    )
                    if col_applied:
                        new_cols[col_applied] = new_data
        if formats:
            for k, v in formats.items():
                func = self._gui._get_user_function(v)
                if callable(func):
                    col_applied, new_data = self.__apply_user_function(
                        func, k if k in cols else None, v, dataframe, "tpf__"
                    )
                    if col_applied:
                        new_cols[col_applied] = new_data
        # deal with dates
        date_cols = col_types[col_types.astype(str).str.startswith("datetime")].index.tolist()
        if len(date_cols) != 0:
            if not is_copied:
                # copy the df so that we don't "mess" with the user's data
                dataframe = dataframe.copy()
            tz = Gui._get_timezone()
            for col in date_cols:
                new_col = _get_date_col_str_name(cols, col)
                re_type = _RE_PD_TYPE.match(str(col_types[col]))
                groups = re_type.groups() if re_type else ()
                if len(groups) > 4 and groups[4]:
                    new_cols[new_col] = (
                        dataframe[col]
                        .dt.tz_convert("UTC")
                        .dt.strftime(_DataAccessor._WS_DATE_FORMAT)
                        .astype(str)
                        .replace("nan", "NaT" if handle_nan else None)
                    )
                else:
                    new_cols[new_col] = (
                        dataframe[col]
                        .dt.tz_localize(tz)
                        .dt.tz_convert("UTC")
                        .dt.strftime(_DataAccessor._WS_DATE_FORMAT)
                        .astype(str)
                        .replace("nan", "NaT" if handle_nan else None)
                    )

            # remove the date columns from the list of columns
            cols = list(set(cols) - set(date_cols))
        if new_cols:
            dataframe = dataframe.assign(**new_cols)
        cols += list(new_cols.keys())
        dataframe = dataframe.loc[:, dataframe.dtypes[dataframe.dtypes.index.astype(str).isin(cols)].index]  # type: ignore[index]
        return dataframe

    def __apply_user_function(
        self,
        user_function: t.Callable,
        column_name: t.Optional[str],
        function_name: str,
        data: pd.DataFrame,
        prefix: t.Optional[str],
    ):
        try:
            new_col_name = f"{prefix}{column_name}__{function_name}" if column_name else function_name
            return new_col_name, data.apply(
                _PandasDataAccessor.__user_function,
                axis=1,
                args=(self._gui, column_name, user_function, function_name),
            )
        except Exception as e:
            _warn(f"Exception raised when invoking user function {function_name}()", e)
        return "", data

    def __format_data(
        self,
        data: pd.DataFrame,
        data_format: _DataFormat,
        orient: str,
        start: t.Optional[int] = None,
        rowcount: t.Optional[int] = None,
        data_extraction: t.Optional[bool] = None,
        handle_nan: t.Optional[bool] = False,
        fullrowcount: t.Optional[int] = None,
    ) -> t.Dict[str, t.Any]:
        ret: t.Dict[str, t.Any] = {
            "format": str(data_format.value),
        }
        if rowcount is not None:
            ret["rowcount"] = rowcount
        if fullrowcount is not None and fullrowcount != rowcount:
            ret["fullrowcount"] = fullrowcount
        if start is not None:
            ret["start"] = start
        if data_extraction is not None:
            ret["dataExtraction"] = data_extraction  # Extract data out of dictionary on front-end
        if data_format is _DataFormat.APACHE_ARROW:
            if not _has_arrow_module:
                raise RuntimeError("Cannot use Arrow as pyarrow package is not installed")
            # Convert from pandas to Arrow
            table = pa.Table.from_pandas(data)  # type: ignore[reportPossiblyUnboundVariable]
            # Create sink buffer stream
            sink = pa.BufferOutputStream()  # type: ignore[reportPossiblyUnboundVariable]
            # Create Stream writer
            writer = pa.ipc.new_stream(sink, table.schema)  # type: ignore[reportPossiblyUnboundVariable]
            # Write data to table
            writer.write_table(table)
            writer.close()
            # End buffer stream
            buf = sink.getvalue()
            # Convert buffer to Python bytes and return
            ret["data"] = buf.to_pybytes()
            ret["orient"] = orient
        else:
            # Workaround for Python built in JSON encoder that does not yet support ignore_nan
            ret["data"] = data.replace([np.nan, pd.NA], [None, None]).to_dict(orient=orient)  # type: ignore
        return ret

    def get_col_types(self, var_name: str, value: t.Any) -> t.Union[None, t.Dict[str, str]]:  # type: ignore
        if isinstance(value, list):
            ret_dict: t.Dict[str, str] = {}
            for i, v in enumerate(value):
                ret_dict.update(
                    {f"{i}/{k}": v for k, v in self.__to_dataframe(v).dtypes.apply(lambda x: x.name.lower()).items()}
                )
            return ret_dict
        return {str(k): v for k, v in self.__to_dataframe(value).dtypes.apply(lambda x: x.name.lower()).items()}

    def __get_data(  # noqa: C901
        self,
        var_name: str,
        df: pd.DataFrame,
        payload: t.Dict[str, t.Any],
        data_format: _DataFormat,
        col_prefix: t.Optional[str] = "",
    ) -> t.Dict[str, t.Any]:
        columns = payload.get("columns", [])
        if col_prefix:
            columns = [c[len(col_prefix) :] if c.startswith(col_prefix) else c for c in columns]
        ret_payload = {"pagekey": payload.get("pagekey", "unknown page")}
        paged = not payload.get("alldata", False)
        is_copied = False

        orig_df = df
        # add index if not chart
        if paged:
            if _PandasDataAccessor.__INDEX_COL not in df.columns:
                is_copied = True
                df = df.assign(**{_PandasDataAccessor.__INDEX_COL: df.index})
            if columns and _PandasDataAccessor.__INDEX_COL not in columns:
                columns.append(_PandasDataAccessor.__INDEX_COL)

        fullrowcount = len(df)
        # filtering
        filters = payload.get("filters")
        if isinstance(filters, list) and len(filters) > 0:
            query = ""
            vars = []
            for fd in filters:
                col = fd.get("col")
                val = fd.get("value")
                action = fd.get("action")
                if isinstance(val, str):
                    if self.__is_date_column(t.cast(pd.DataFrame, df), col):
                        val = datetime.fromisoformat(val[:-1])
                    vars.append(val)
                val = f"@vars[{len(vars) - 1}]" if isinstance(val, (str, datetime)) else val
                right = f".str.contains({val})" if action == "contains" else f" {action} {val}"
                if query:
                    query += " and "
                query += f"`{col}`{right}"
            try:
                df = df.query(query)
                is_copied = True
            except Exception as e:
                _warn(f"Dataframe filtering: invalid query '{query}' on {df.head()}", e)

        dict_ret: t.Optional[t.Dict[str, t.Any]]
        if paged:
            aggregates = payload.get("aggregates")
            applies = payload.get("applies")
            if isinstance(aggregates, list) and len(aggregates) and isinstance(applies, dict):
                applies_with_fn = {
                    k: v if v in _PandasDataAccessor.__AGGREGATE_FUNCTIONS else self._gui._get_user_function(v)
                    for k, v in applies.items()
                }

                for col in columns:
                    if col not in applies_with_fn.keys():
                        applies_with_fn[col] = "first"
                try:
                    df = t.cast(pd.DataFrame, df).groupby(aggregates).agg(applies_with_fn)
                except Exception:
                    _warn(f"Cannot aggregate {var_name} with groupby {aggregates} and aggregates {applies}.")
            inf = payload.get("infinite")
            if inf is not None:
                ret_payload["infinite"] = inf
            # real number of rows is needed to calculate the number of pages
            rowcount = len(df)
            # here we'll deal with start and end values from payload if present
            if isinstance(payload.get("start", 0), int):
                start = int(payload.get("start", 0))
            else:
                try:
                    start = int(str(payload["start"]), base=10)
                except Exception:
                    _warn(f'start should be an int value {payload["start"]}.')
                    start = 0
            if isinstance(payload.get("end", -1), int):
                end = int(payload.get("end", -1))
            else:
                try:
                    end = int(str(payload["end"]), base=10)
                except Exception:
                    end = -1
            if start < 0 or start >= rowcount:
                start = 0
            if end < 0 or end >= rowcount:
                end = rowcount - 1
            if payload.get("reverse", False):
                diff = end - start
                end = rowcount - 1 - start
                if end < 0:
                    end = rowcount - 1
                start = end - diff
                if start < 0:
                    start = 0
            # deal with sort
            order_by = payload.get("orderby")
            if isinstance(order_by, str) and len(order_by):
                try:
                    if df.columns.dtype.name == "int64":
                        order_by = int(order_by)
                    new_indexes = t.cast(pd.DataFrame, df)[order_by].values.argsort(axis=0)
                    if payload.get("sort") == "desc":
                        # reverse order
                        new_indexes = new_indexes[::-1]
                    new_indexes = new_indexes[slice(start, end + 1)]
                except Exception:
                    _warn(f"Cannot sort {var_name} on columns {order_by}.")
                    new_indexes = slice(start, end + 1)  # type: ignore
            else:
                new_indexes = slice(start, end + 1)  # type: ignore
            df = self.__build_transferred_cols(
                columns,
                t.cast(pd.DataFrame, df),
                styles=payload.get("styles"),
                tooltips=payload.get("tooltips"),
                is_copied=is_copied,
                new_indexes=t.cast(np.ndarray, new_indexes),
                handle_nan=payload.get("handlenan", False),
                formats=payload.get("formats"),
            )
            dict_ret = self.__format_data(
                df,
                data_format,
                "records",
                start,
                rowcount,
                handle_nan=payload.get("handlenan", False),
                fullrowcount=fullrowcount,
            )
            compare = payload.get("compare")
            if isinstance(compare, str):
                comp_df = _compare_function(
                    self._gui, compare, var_name, t.cast(pd.DataFrame, orig_df), payload.get("compare_datas", "")
                )
                if isinstance(comp_df, pd.DataFrame) and not comp_df.empty:
                    try:
                        if isinstance(comp_df.columns[0], tuple):
                            cols: t.List[t.Hashable] = [c for c in comp_df.columns if c[1] == "other"]
                            comp_df = t.cast(pd.DataFrame, comp_df.get(cols))
                            comp_df.columns = t.cast(pd.Index, [t.cast(tuple, c)[0] for c in cols])
                        comp_df.dropna(axis=1, how="all", inplace=True)
                        comp_df = self.__build_transferred_cols(
                            columns, comp_df, new_indexes=t.cast(np.ndarray, new_indexes)
                        )
                        dict_ret["comp"] = self.__format_data(comp_df, data_format, "records").get("data")
                    except Exception as e:
                        _warn("Pandas accessor compare raised an exception", e)

        else:
            ret_payload["alldata"] = True
            decimator_payload: t.Dict[str, t.Any] = payload.get("decimatorPayload", {})
            decimators = decimator_payload.get("decimators", [])
            decimated_dfs: t.List[pd.DataFrame] = []
            for decimator_pl in decimators:
                if decimator_pl is None:
                    continue
                decimator = decimator_pl.get("decimator")
                if decimator is None:
                    x_column = decimator_pl.get("xAxis", "")
                    y_column = decimator_pl.get("yAxis", "")
                    z_column = decimator_pl.get("zAxis", "")
                    filtered_columns = [x_column, y_column, z_column] if z_column else [x_column, y_column]
                    decimated_df = df.copy().filter(filtered_columns, axis=1)
                    decimated_dfs.append(decimated_df)
                    continue
                decimator_instance = (
                    self._gui._get_user_instance(decimator, PropertyType.decimator.value)
                    if decimator is not None
                    else None
                )
                if isinstance(decimator_instance, PropertyType.decimator.value):
                    # Run the on_decimate method -> check if the decimator should be applied
                    # -> apply the decimator
                    decimated_df, is_decimator_applied, is_copied = decimator_instance._on_decimate(
                        df, decimator_pl, decimator_payload, is_copied
                    )
                    # add decimated dataframe to the list of decimated
                    decimated_dfs.append(decimated_df)
                    if is_decimator_applied:
                        self._gui._call_on_change(f"{var_name}.{decimator}.nb_rows", len(decimated_df))
            # merge the decimated dataFrames
            if len(decimated_dfs) > 1:
                # get the unique columns from all decimated dataFrames
                decimated_columns = pd.Index([])
                for _df in decimated_dfs:
                    decimated_columns = decimated_columns.append(_df.columns)
                # find the columns that are duplicated across dataFrames
                overlapping_columns = decimated_columns[decimated_columns.duplicated()].unique()
                # concatenate the dataFrames without overwriting columns
                merged_df = pd.concat(decimated_dfs, axis=1)
                # resolve overlapping columns by combining values
                for col in overlapping_columns:
                    # for each overlapping column, combine the values across dataFrames
                    # (e.g., take the first non-null value)
                    cols_to_combine = merged_df.loc[:, col].columns
                    merged_df[col] = merged_df[cols_to_combine].bfill(axis=1).iloc[:, 0]
                # drop duplicated col since they are now the same
                df = merged_df.loc[:,~merged_df.columns.duplicated()]
            elif len(decimated_dfs) == 1:
                df = decimated_dfs[0]
            if data_format is _DataFormat.CSV:
                df = self.__build_transferred_cols(
                    columns,
                    t.cast(pd.DataFrame, df),
                    is_copied=is_copied,
                    handle_nan=payload.get("handlenan", False),
                )
                ret_payload["df"] = df
                dict_ret = None
            else:
                df = self.__build_transferred_cols(
                    columns,
                    t.cast(pd.DataFrame, df),
                    styles=payload.get("styles"),
                    tooltips=payload.get("tooltips"),
                    is_copied=is_copied,
                    handle_nan=payload.get("handlenan", False),
                    formats=payload.get("formats"),
                )
                dict_ret = self.__format_data(df, data_format, "list", data_extraction=True)

        ret_payload["value"] = dict_ret
        return ret_payload

    def get_data(
        self, var_name: str, value: t.Any, payload: t.Dict[str, t.Any], data_format: _DataFormat
    ) -> t.Dict[str, t.Any]:
        if isinstance(value, list):
            # If is_chart data
            if payload.get("alldata", False):
                ret_payload = {
                    "alldata": True,
                    "value": {"multi": True},
                    "pagekey": payload.get("pagekey", "unknown page"),
                }
                data = []
                for i, v in enumerate(value):
                    ret = (
                        self.__get_data(var_name, self.__to_dataframe(v), payload, data_format, f"{i}/")
                        if isinstance(v, _PandasDataAccessor.__types)
                        else {}
                    )
                    ret_val = ret.get("value", {})
                    data.append(ret_val.pop("data", None))
                    ret_payload.get("value", {}).update(ret_val)
                ret_payload["value"]["data"] = data
                return ret_payload
            else:
                value = value[0]
        return self.__get_data(var_name, self.__to_dataframe(value), payload, data_format)

    def on_edit(self, value: t.Any, payload: t.Dict[str, t.Any]):
        df = self.to_pandas(value)
        if not isinstance(df, pd.DataFrame):
            raise ValueError(f"Cannot edit {type(value)}.")
        df.at[payload["index"], payload["col"]] = payload["value"]
        return self._from_pandas(df, type(value))

    def on_delete(self, value: t.Any, payload: t.Dict[str, t.Any]):
        df = self.to_pandas(value)
        if not isinstance(df, pd.DataFrame):
            raise ValueError(f"Cannot delete a row from {type(value)}.")
        return self._from_pandas(df.drop(payload["index"]), type(value))

    def on_add(self, value: t.Any, payload: t.Dict[str, t.Any], new_row: t.Optional[t.List[t.Any]] = None):
        df = self.to_pandas(value)
        if not isinstance(df, pd.DataFrame):
            raise ValueError(f"Cannot add a row to {type(value)}.")
        # Save the insertion index
        index = payload["index"]
        # Create the new row (Column value types must match the original DataFrame's)
        col_types = self.get_col_types("", df)
        if col_types:
            new_row = [0 if is_numeric_dtype(df[c]) else "" for c in list(col_types)] if new_row is None else new_row
            if index > 0:
                # Column names and value types must match the original DataFrame
                new_df = pd.DataFrame([new_row], columns=list(col_types))
                # Split the DataFrame
                rows_before = df.iloc[:index]
                rows_after = df.iloc[index:]
                return self._from_pandas(pd.concat([rows_before, new_df, rows_after], ignore_index=True), type(value))
            else:
                df = df.copy()
                # Insert as the new first row
                df.loc[-1] = new_row  # Insert the new row
                df.index = df.index + 1  # Shift index
                return self._from_pandas(df.sort_index(), type(value))
        return value

    def to_csv(self, var_name: str, value: t.Any):
        df = self.to_pandas(value)
        if not isinstance(df, pd.DataFrame):
            raise ValueError(f"Cannot export {type(value)} to csv.")
        dict_ret = self.__get_data(var_name, df, {"alldata": True}, _DataFormat.CSV)
        if isinstance(dict_ret, dict):
            dfr = dict_ret.get("df")
            if isinstance(dfr, pd.DataFrame):
                fd, temp_path = mkstemp(".csv", var_name, text=True)
                with os.fdopen(fd, "wt", newline="") as csv_file:
                    dfr.to_csv(csv_file, index=False)

                return temp_path
        return None



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import re
import typing as t
from abc import ABC, abstractmethod
from os import path

from charset_normalizer import detect

from taipy.common.logger._taipy_logger import _TaipyLogger

from ..page import Page
from ..utils import _is_in_notebook, _varname_from_content
from ._html import _TaipyHTMLParser

if t.TYPE_CHECKING:
    from watchdog.observers.api import BaseObserver

    from ..gui import Gui


class _Renderer(Page, ABC):
    def __init__(self, **kwargs) -> None:
        """NOT DOCUMENTED
        Initialize a new _Renderer with the indicated content.

        Arguments:
            content (str): The text content or the path to the file holding the text to be transformed.

        If *content* is a path to a readable file, the file is read entirely as the text template.
        """
        from ..builder._element import _Element  # noqa: F811

        super().__init__(**kwargs)
        content: t.Optional[t.Union[str, _Element]] = kwargs.get("content", None)
        if content is None:
            raise ValueError("'content' argument is missing for class '_Renderer'")
        self._content = ""
        self._base_element: t.Optional[_Element] = None
        self._filepath = ""
        self._observer: t.Optional["BaseObserver"] = None
        self._encoding: t.Optional[str] = kwargs.get("encoding", None)
        if isinstance(content, str):
            self.__process_content(content)
        elif isinstance(content, _Element):
            self._base_element = content
        else:
            raise ValueError(
                f"'content' argument has incorrect type '{type(content).__name__}'. This must be a string or an Builder element."  # noqa: E501
            )

    def __process_content(self, content: str) -> None:
        relative_file_path = (
            None if self._frame is None else path.join(path.dirname(self._frame.f_code.co_filename), content)
        )
        if relative_file_path is not None and path.exists(relative_file_path) and path.isfile(relative_file_path):
            content = relative_file_path
        if content == relative_file_path or (path.exists(content) and path.isfile(content)):
            self.__parse_file_content(content)
            # Watchdog observer: watch for file changes
            if _is_in_notebook() and self._observer is None:
                self.__observe_file_change(content)
            return
        self._content = self.__sanitize_content(content)

    def __observe_file_change(self, file_path: str):
        from watchdog.observers import Observer

        from .utils import FileWatchdogHandler

        self._observer = Observer()
        file_path = path.abspath(file_path)
        self._observer.schedule(FileWatchdogHandler(file_path, self), path.dirname(file_path), recursive=False)
        self._observer.start()

    def __parse_file_content(self, content):
        with open(t.cast(str, content), "rb") as f:
            file_content = f.read()
            encoding = "utf-8"
            if self._encoding is not None:
                encoding = self._encoding
                _TaipyLogger._get_logger().info(f"'{encoding}' encoding was used to decode file '{content}'.")
            elif (detected_encoding := detect(file_content)["encoding"]) is not None:
                encoding = detected_encoding
                _TaipyLogger._get_logger().info(f"Detected '{encoding}' encoding for file '{content}'.")
            else:
                _TaipyLogger._get_logger().info(f"Using default '{encoding}' encoding for file '{content}'.")
            self._content = self.__sanitize_content(file_content.decode(encoding))
            # Save file path for error handling
            self._filepath = content

    def __sanitize_content(self, content: str) -> str:
        # Replace all CRLF (\r\n) and CR (\r) by LF (\n)
        return re.sub(r"\r", "\n", re.sub(r"\r\n", "\n", content))

    def set_content(self, content: str) -> None:
        if not _is_in_notebook():
            raise RuntimeError("'set_content()' must be used in an IPython notebook context")
        self.__process_content(content)
        if self._notebook_gui is not None and self._notebook_page is not None:
            if self._notebook_gui._config.root_page is self._notebook_page:
                self._notebook_gui._navigate("/", {"tp_reload_all": "true"})
                return
            self._notebook_gui._navigate(self._notebook_page._route, {"tp_reload_same_route_only": "true"})

    def _get_content_detail(self, gui: "Gui") -> str:
        if self._filepath:
            return f"in file '{self._filepath}'"
        if varname := _varname_from_content(gui, self._content):
            return f"in variable '{varname}'"
        return ""

    @abstractmethod
    def render(self, gui: "Gui") -> str:
        pass


class _EmptyPage(_Renderer):
    def __init__(self) -> None:
        super().__init__(content="<PageContent />")

    def render(self, gui: "Gui") -> str:
        return self._content


class Markdown(_Renderer):
    """Page generator for *Markdown* text.

    Taipy can use Markdown text to create pages that are the base of
    user interfaces.

    You can find details on the Taipy Markdown-specific syntax and how to add
    Taipy Visual Elements in the [section on Markdown](../../../../../userman/gui/pages/markdown.md)
    of the User Manual.
    """

    def __init__(self, content: str, **kwargs) -> None:
        """Initialize a new `Markdown` page.

        Arguments:
            content (str): The text content or the path to the file holding the Markdown text
                to be transformed.<br/>
                If *content* is a path to a readable file, the file is read as the Markdown
                template content.

        The `Markdown` constructor supports the *style* parameter as explained in the
        [section on Styling](../../../../../userman/gui/styling/index.md#style-sheets) and in the
        `(taipy.gui.Page.)set_style()^` method.
        """
        kwargs["content"] = content
        super().__init__(**kwargs)

    # Generate JSX from Markdown
    def render(self, gui: "Gui") -> str:
        return gui._markdown.convert(self._content)


class Html(_Renderer):
    """Page generator for *HTML* text.

    Taipy can use HTML code to create pages that are the base of
    user interfaces.

    You can find details on HTML-specific constructs and how to add
    Taipy Visual Elements in the [section on HTML](../../../../../userman/gui/pages/html.md)
    of the User Manual.
    """

    def __init__(self, content: str, **kwargs) -> None:
        """Initialize a new `Html` page.

        Arguments:
            content (str): The text content or the path to the file holding the HTML text to
                be transformed.<br/>
                If *content* is a path to a readable file, the file is read as the HTML
                template content.

        The `Html` constructor supports the *style* parameter as explained in the
        [section on Styling](../../../../../userman/gui/styling/index.md#style-sheets) and in the
        `(taipy.gui.Page.)set_style()^` method.
        """
        kwargs["content"] = content
        super().__init__(**kwargs)
        self.head = None

    # Modify path routes
    def modify_taipy_base_url(self, base_url):
        self._content = self._content.replace("{{taipy_base_url}}", f"{base_url}")

    # Generate JSX from HTML
    def render(self, gui: "Gui") -> str:
        parser = _TaipyHTMLParser(gui)
        parser.feed_data(self._content)
        self.head = parser.head
        return parser.get_jsx()



# Copyright 2021-2025 Avaiga Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import contextlib
import typing as t

from flask import has_request_context
from flask.globals import request

from ..server import _Server
from ._page import ResourceHandler, _ExternalResourceHandlerManager


def is_in_custom_page_context() -> bool:
    "NOT DOCUMENTED"
    resource_handler_id = None
    with contextlib.suppress(Exception):
        if has_request_context():
            resource_handler_id = request.cookies.get(_Server._RESOURCE_HANDLER_ARG, None)
    return resource_handler_id is not None


def get_current_resource_handler() -> t.Optional[ResourceHandler]:
    "NOT DOCUMENTED"
    resource_handler_id = None
    with contextlib.suppress(Exception):
        if has_request_context():
            resource_handler_id = request.cookies.get(_Server._RESOURCE_HANDLER_ARG, None)
    return _ExternalResourceHandlerManager().get(resource_handler_id) if resource_handler_id else None
